import hashlib
import importlib
import string
from inspect import getfullargspec, isclass

import attr
from faker import Faker
from faker.providers import BaseProvider


def generate_faker_seed(value):
    """Generate a seed value for faker."""
    if not isinstance(value, str):
        value = str(value)

    h = hashlib.new("md5")
    h.update(value.encode("utf-8"))
    return int(h.hexdigest()[:16], 16)


class TestProvider(BaseProvider):
    """A demo faker provider for testing string providers"""

    def foo(self):
        return "foo"


class FakerFormatter(string.Formatter):
    """A formatter that can get values from Faker generators."""

    def format_field(self, obj, format_spec):
        """

        :param obj: A faker Faker
        :param format_spec: a generator
        :return: A string generated by
        """
        generator = format_spec
        kwargs = {}
        if "|" in format_spec:
            try:
                newgenerator, potential_kwargs = format_spec.split("|")
                for part in potential_kwargs.split(","):
                    k, v = part.split("=")
                    if v == "None":
                        v = None
                    elif v == "True":
                        v = True
                    elif v == "False":
                        v = False
                    elif v.isdigit():
                        v = int(v)
                    kwargs[k] = v
                generator = newgenerator
            except ValueError:
                # If more than one "|"  don't try to parse
                # If the kwargs aren't of form x=y then don't try to parse
                pass

        value = None
        if callable(getattr(obj, generator)):
            c = getattr(obj, generator)
            argspec = getfullargspec(c)
            if len(argspec.args) == 1:
                value = getattr(obj, generator)()
            elif kwargs:
                value = getattr(obj, generator)(**kwargs)
            else:
                value = c

        if value is not None and not isinstance(value, str):
            value = str(value)
        return value or "Unknown fake generator"


@attr.s
class FakerAnonymizer:
    """Returns a deterministically generated fake value that depends on the
    input value."""

    format_str = attr.ib()
    locale = attr.ib(default="en_US")
    postprocessor = attr.ib(default=None)
    providers = attr.ib(default=None)

    def __attrs_post_init__(self):
        self.fake = Faker(self.locale)
        self.formatter = FakerFormatter()
        for p in self._clean_providers(self.providers):
            self.fake.add_provider(p)

    def _clean_providers(self, providers):
        """Convert a list of anonymizer providers into classes suitable for
        adding with faker.add_provider"""
        if not providers:
            return []

        if not isinstance(providers, (list, tuple)):
            providers = [providers]

        cleaned_providers = []
        for provider in providers:
            if isinstance(provider, str):
                # dynamically import the provider
                parts = provider.split(".")
                if len(parts) > 1:
                    _module = ".".join(parts[:-1])
                    _provider_class = parts[-1]
                    try:
                        _mod = importlib.import_module(_module)
                        _provider = getattr(_mod, _provider_class, None)
                        if _provider is None:
                            # TODO: log an issue, provider not found in module
                            continue
                        elif not issubclass(_provider, BaseProvider):
                            # TODO: log an issue, provider not generator
                            continue
                        else:
                            cleaned_providers.append(_provider)

                    except ImportError:
                        # TODO: log an issue, can't import module
                        continue
            elif isclass(provider) and issubclass(provider, BaseProvider):
                cleaned_providers.append(provider)
            else:
                # TODO: log an issue, provider is not an importable string
                #  or a ProviderBase
                continue

        return cleaned_providers

    def __call__(self, value):
        self.fake.seed_instance(generate_faker_seed(value))
        value = self.formatter.format(self.format_str, fake=self.fake)
        if self.postprocessor is None:
            return value
        else:
            return self.postprocessor(value)



from copy import copy

import pytest
from sqlalchemy import func, join
from tests.test_base import RecipeTestCase

from recipe import BadRecipe, Dimension, Filter, Having, Metric, Recipe, Shelf


class TestRecipeIngredients(RecipeTestCase):
    def test_dimension(self):
        recipe = self.recipe().metrics("age").dimensions("first")
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.first AS first,
       sum(foo.age) AS age
FROM foo
GROUP BY first""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            first,age,first_id
            hi,15,hi
            """,
        )

    def test_raw_metrics_raw_dimensions(self):
        """Metric_ids and dimension_ids hold unique used metrics and dimensions.
        raw_metrics and raw_dimensions store ingredients as added."""

        # We call dimensions and metrics multiple times
        recipe = (
            self.recipe().metrics("age").dimensions("first", "first").metrics("age")
        )
        assert (
            recipe.to_sql()
            == """SELECT foo.first AS first,
       sum(foo.age) AS age
FROM foo
GROUP BY first"""
        )
        self.assertRecipeCSV(
            recipe,
            """
            first,age,first_id
            hi,15,hi
            """,
        )
        self.assertEqual(recipe.metric_ids, ("age",))
        self.assertEqual(recipe.dimension_ids, ("first",))
        self.assertEqual(recipe.raw_dimensions, ("first", "first"))
        self.assertEqual(recipe.raw_metrics, ("age", "age"))

    def test_idvaluedimension(self):
        recipe = self.recipe().metrics("age").dimensions("firstlast")
        assert (
            recipe.to_sql()
            == """SELECT foo.first AS firstlast_id,
       foo.last AS firstlast,
       sum(foo.age) AS age
FROM foo
GROUP BY firstlast_id,
         firstlast"""
        )
        self.assertRecipeCSV(
            recipe,
            """
            firstlast_id,firstlast,age
            hi,fred,10
            hi,there,5
            """,
        )

    def test_multirole_dimension(self):
        """Create a dimension with extra roles and lookup"""
        d = Dimension(
            self.basic_table.c.last,
            id_expression=self.basic_table.c.first,
            age_expression=self.basic_table.c.age,
            id="d",
        )
        recipe = self.recipe().metrics("age").dimensions(d)
        assert (
            recipe.to_sql()
            == """SELECT foo.first AS d_id,
       foo.last AS d,
       foo.age AS d_age,
       sum(foo.age) AS age
FROM foo
GROUP BY d_id,
         d,
         d_age"""
        )
        self.assertRecipeCSV(
            recipe,
            """
            d_id,d,d_age,age
            hi,fred,10,10
            hi,there,5,5
            """,
        )

    def test_multirole_dimension_with_lookup(self):
        """Create a dimension with extra roles and lookup"""
        d = Dimension(
            self.basic_table.c.last,
            id_expression=self.basic_table.c.first,
            age_expression=self.basic_table.c.age,
            id="d",
            lookup={},
            lookup_default="DEFAULT",
        )
        recipe = self.recipe().metrics("age").dimensions(d)

        assert (
            recipe.to_sql()
            == """SELECT foo.first AS d_id,
       foo.last AS d_raw,
       foo.age AS d_age,
       sum(foo.age) AS age
FROM foo
GROUP BY d_id,
         d_raw,
         d_age"""
        )
        self.assertRecipeCSV(
            recipe,
            """
            d_id,d_raw,d_age,age,d
            hi,fred,10,10,DEFAULT
            hi,there,5,5,DEFAULT
            """,
        )

    def test_offset(self):
        recipe = self.recipe().metrics("age").dimensions("first").offset(1)
        assert (
            recipe.to_sql()
            == """SELECT foo.first AS first,
       sum(foo.age) AS age
FROM foo
GROUP BY first
LIMIT ?
OFFSET 1"""
        )
        self.assertEqual(len(recipe.all()), 0)
        self.assertEqual(recipe.stats.rows, 0)
        self.assertEqual(recipe.one(), [])

    def test_is_postgres(self):
        recipe = self.recipe().metrics("age")
        self.assertFalse(recipe._is_postgres())

    def test_use_cache(self):
        recipe = self.recipe().metrics("age")
        self.assertTrue(recipe._use_cache)

        recipe = self.recipe().metrics("age").use_cache(False)
        self.assertFalse(recipe._use_cache)

        with self.assertRaises(AssertionError):
            self.recipe().metrics("age").use_cache("potatoe")

    def test_cache_region(self):
        recipe = self.recipe().metrics("age")
        self.assertEqual(recipe._cache_region, "default")

        recipe = self.recipe().metrics("age").cache_region("foo")
        self.assertEqual(recipe._cache_region, "foo")

        with self.assertRaises(AssertionError):
            self.recipe().metrics("age").cache_region(22)

    def test_cache_prefix(self):
        recipe = self.recipe().metrics("age")
        self.assertEqual(recipe._cache_prefix, "default")

        recipe = self.recipe().metrics("age").cache_prefix("foo")
        self.assertEqual(recipe._cache_prefix, "foo")

        with self.assertRaises(AssertionError):
            self.recipe().metrics("age").cache_prefix(22)

    def test_dataset(self):
        recipe = self.recipe().metrics("age").dimensions("first")
        assert recipe.dataset.json == '[{"first": "hi", "age": 15, "first_id": "hi"}]'

        # Line delimiter is \r\n
        self.assertEqual(recipe.dataset.csv, "first,age,first_id\r\nhi,15,hi\r\n")

    def test_session(self):
        recipe = self.recipe().metrics("age").dimensions("first")
        assert (
            recipe.to_sql()
            == """SELECT foo.first AS first,
       sum(foo.age) AS age
FROM foo
GROUP BY first"""
        )
        self.assertEqual(recipe.all()[0].first, "hi")
        self.assertEqual(recipe.all()[0].age, 15)
        self.assertEqual(recipe.stats.rows, 1)

        sess = self.oven.Session()
        recipe.reset()
        recipe.session(sess)

    def test_shelf(self):
        recipe = self.recipe().metrics("age").dimensions("first")
        self.assertEqual(len(recipe._shelf), 4)
        recipe.shelf(None)
        self.assertEqual(len(recipe._shelf), 0)
        recipe.shelf(self.shelf)
        self.assertEqual(len(recipe._shelf), 4)
        recipe.shelf({})
        self.assertEqual(len(recipe._shelf), 0)
        with self.assertRaises(BadRecipe):
            recipe.shelf(52)

    def test_dimension2(self):
        recipe = self.recipe().metrics("age").dimensions("last").order_by("last")
        assert (
            recipe.to_sql()
            == """SELECT foo.last AS last,
       sum(foo.age) AS age
FROM foo
GROUP BY last
ORDER BY last"""
        )
        self.assertRecipeCSV(
            recipe,
            """
            last,age,last_id
            fred,10,fred
            there,5,there
            """,
        )

    def test_order_bys(self):
        recipe = self.recipe().metrics("age").dimensions("last").order_by("last")
        assert (
            recipe.to_sql()
            == """SELECT foo.last AS last,
       sum(foo.age) AS age
FROM foo
GROUP BY last
ORDER BY last"""
        )
        self.assertRecipeCSV(
            recipe,
            """
            last,age,last_id
            fred,10,fred
            there,5,there
            """,
        )

        recipe = self.recipe().metrics("age").dimensions("last").order_by("age")
        assert (
            recipe.to_sql()
            == """SELECT foo.last AS last,
       sum(foo.age) AS age
FROM foo
GROUP BY last
ORDER BY age"""
        )
        self.assertRecipeCSV(
            recipe,
            """
            last,age,last_id
            there,5,there
            fred,10,fred
            """,
        )

        recipe = self.recipe().metrics("age").dimensions("last").order_by("-age")
        assert (
            recipe.to_sql()
            == """SELECT foo.last AS last,
       sum(foo.age) AS age
FROM foo
GROUP BY last
ORDER BY age DESC"""
        )
        self.assertRecipeCSV(
            recipe,
            """
            last,age,last_id
            fred,10,fred
            there,5,there
            """,
        )

        # Idvalue dimension
        recipe = (
            self.recipe().metrics("age").dimensions("firstlast").order_by("firstlast")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.first AS firstlast_id,
       foo.last AS firstlast,
       sum(foo.age) AS age
FROM foo
GROUP BY firstlast_id,
         firstlast
ORDER BY firstlast,
         firstlast_id""",
        )
        recipe = (
            self.recipe().metrics("age").dimensions("firstlast").order_by("-firstlast")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.first AS firstlast_id,
       foo.last AS firstlast,
       sum(foo.age) AS age
FROM foo
GROUP BY firstlast_id,
         firstlast
ORDER BY firstlast DESC,
         firstlast_id DESC""",
        )

        # Dimensions can define their own ordering
        d = Dimension(
            self.basic_table.c.last,
            id="d",
            id_expression=self.basic_table.c.first,
            order_by_expression=func.upper(self.basic_table.c.last),
        )
        recipe = self.recipe().metrics("age").dimensions(d).order_by(d)
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.first AS d_id,
       foo.last AS d,
       upper(foo.last) AS d_order_by,
       sum(foo.age) AS age
FROM foo
GROUP BY d_id,
         d,
         d_order_by
ORDER BY d_order_by,
         d,
         d_id""",
        )

    def test_recipe_init(self):
        """Test that all options can be passed in the init"""
        recipe = self.recipe(metrics=("age",), dimensions=("last",)).order_by("last")
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.last AS last,
       sum(foo.age) AS age
FROM foo
GROUP BY last
ORDER BY last""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            last,age,last_id
            fred,10,fred
            there,5,there
            """,
        )

        recipe = self.recipe(
            metrics=["age"],
            dimensions=["first"],
            filters=[self.basic_table.c.age > 4],
            order_by=["first"],
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.first AS first,
       sum(foo.age) AS age
FROM foo
WHERE foo.age > 4
GROUP BY first
ORDER BY first""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            first,age,first_id
            hi,15,hi
            """,
        )

    def test_recipe_filters_ordered(self):
        """Filters appear in the SQL in order and are deduped"""
        recipe = (
            self.recipe()
            .metrics("age")
            .dimensions("last")
            .filters(
                Filter(self.basic_table.c.age > 3),
                self.basic_table.c.age > 4,
                self.basic_table.c.age > 5,
                Filter(self.basic_table.c.age > 4),
                self.basic_table.c.age > 4,
            )
            .order_by("last")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.last AS last,
       sum(foo.age) AS age
FROM foo
WHERE foo.age > 3
  AND foo.age > 4
  AND foo.age > 5
GROUP BY last
ORDER BY last""",
        )

        recipe = self.recipe(
            metrics=["age"],
            dimensions=["first"],
            filters=[
                self.basic_table.c.age > 5,
                self.basic_table.c.age > 4,
                self.basic_table.c.age > 3,
            ],
            order_by=["first"],
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.first AS first,
       sum(foo.age) AS age
FROM foo
WHERE foo.age > 3
  AND foo.age > 4
  AND foo.age > 5
GROUP BY first
ORDER BY first""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            first,age,first_id
            hi,10,hi
            """,
        )

    def test_from_config(self):
        shelf = copy(self.mytable_shelf)
        shelf["ageover4"] = Filter(self.basic_table.c.age > 4)
        config = {
            "dimensions": ["first", "last"],
            "metrics": ["age"],
            "filters": ["ageover4"],
        }
        recipe = self.recipe_from_config(config, shelf=shelf)
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.first AS first,
       foo.last AS last,
       sum(foo.age) AS age
FROM foo
WHERE foo.age > 4
GROUP BY first,
         last""",
        )

    def test_order_bys_not_matching_ingredients(self):
        """If an order_by is not found in dimensions+metrics, we ignore it"""
        recipe = self.recipe().metrics("age").dimensions("first").order_by("last")
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.first AS first,
       sum(foo.age) AS age
FROM foo
GROUP BY first""",
        )
        self.assertRecipeSQLNotContains(recipe, "last")
        self.assertEqual(recipe.all()[0].first, "hi")
        self.assertEqual(recipe.all()[0].age, 15)
        self.assertEqual(recipe.stats.rows, 1)

    def test_from_config_extra_kwargs(self):
        config = {"dimensions": ["last"], "metrics": ["age"]}
        recipe = Recipe.from_config(self.shelf, config, order_by=["last"]).session(
            self.session
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.last AS last,
       sum(foo.age) AS age
FROM foo
GROUP BY last
ORDER BY last""",
        )

    def test_recipe_empty(self):
        recipe = self.recipe()
        with self.assertRaises(BadRecipe):
            recipe.all()

    def test_recipe_multi_tables(self):
        """All ingredients in the recipe must use the same table"""
        dim = Dimension(self.scores_table.c.username)
        recipe = self.recipe().dimensions("last", dim).metrics("age")
        with self.assertRaises(BadRecipe):
            recipe.all()

    def test_recipe_as_table(self):
        recipe = self.recipe().dimensions("last").metrics("age")
        tbl = recipe.as_table()
        self.assertEqual(tbl.name, recipe._id)

        tbl = recipe.as_table(name="foo")
        self.assertEqual(tbl.name, "foo")

    def test_filter(self):
        recipe = (
            self.recipe()
            .metrics("age")
            .dimensions("last")
            .filters(self.basic_table.c.age > 2)
            .order_by("last")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.last AS last,
       sum(foo.age) AS age
FROM foo
WHERE foo.age > 2
GROUP BY last
ORDER BY last""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            last,age,last_id
            fred,10,fred
            there,5,there
            """,
        )

    def test_having(self):
        hv = Having(func.sum(self.basic_table.c.age) < 10)
        recipe = (
            self.recipe()
            .metrics("age")
            .dimensions("last")
            .filters(self.basic_table.c.age > 2)
            .filters(hv)
            .order_by("last")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.last AS last,
       sum(foo.age) AS age
FROM foo
WHERE foo.age > 2
GROUP BY last
HAVING sum(foo.age) < 10
ORDER BY last""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            last,age,last_id
            there,5,there
            """,
        )


class StatsTestCase(RecipeTestCase):
    def test_stats(self):
        recipe = self.recipe().metrics("age").dimensions("last")
        recipe.all()

        # Stats are ready after the recipe is run
        self.assertEqual(recipe.stats.rows, 2)
        self.assertLess(recipe.stats.dbtime, 1.0)
        self.assertLess(recipe.stats.enchanttime, 1.0)
        self.assertFalse(recipe.stats.from_cache)


class NestedRecipeTestCase(RecipeTestCase):
    def test_nested_recipe(self):
        recipe = self.recipe().metrics("age").dimensions("last")
        from recipe import Metric, Dimension

        subq = recipe.subquery(name="anon")
        nested_shelf = Shelf(
            {"age": Metric(func.sum(subq.c.age)), "last": Dimension(subq.c.last)}
        )
        r = self.recipe(shelf=nested_shelf).dimensions("last").metrics("age")
        self.assertRecipeSQL(
            r,
            """SELECT anon.last AS last,
       sum(anon.age) AS age
FROM
  (SELECT foo.last AS last,
          sum(foo.age) AS age
   FROM foo
   GROUP BY last) AS anon
GROUP BY last""",
        )
        self.assertEqual(len(r.all()), 2)


class SelectFromTestCase(RecipeTestCase):
    def test_recipe_with_select_from(self):
        j = join(
            self.census_table,
            self.state_fact_table,
            self.census_table.c.state == self.state_fact_table.c.name,
        )

        from recipe import Dimension, Metric

        shelf = Shelf(
            {
                "region": Dimension(self.state_fact_table.c.census_region_name),
                "pop": Metric(func.sum(self.census_table.c.pop2000)),
            }
        )

        r = (
            Recipe(shelf=shelf, session=self.session)
            .dimensions("region")
            .metrics("pop")
            .select_from(j)
        )

        self.assertRecipeSQL(
            r,
            """SELECT state_fact.census_region_name AS region,
       sum(census.pop2000) AS pop
FROM census
JOIN state_fact ON census.state = state_fact.name
GROUP BY region""",
        )
        self.assertRecipeCSV(
            r,
            """
            region,pop,region_id
            Northeast,609480,Northeast
            South,5685230,South
            """,
        )


class ShelfSelectFromTestCase(RecipeTestCase):
    def test_recipe_with_select_from(self):
        from recipe import Dimension, Metric

        shelf = Shelf(
            {
                "region": Dimension(self.state_fact_table.c.census_region_name),
                "pop": Metric(func.sum(self.census_table.c.pop2000)),
            },
            select_from=join(
                self.census_table,
                self.state_fact_table,
                self.census_table.c.state == self.state_fact_table.c.name,
            ),
        )

        self.assertTrue(shelf.Meta.select_from is not None)

        r = (
            Recipe(shelf=shelf, session=self.session)
            .dimensions("region")
            .metrics("pop")
        )
        self.assertTrue(r._select_from is not None)

        self.assertRecipeSQL(
            r,
            """SELECT state_fact.census_region_name AS region,
       sum(census.pop2000) AS pop
FROM census
JOIN state_fact ON census.state = state_fact.name
GROUP BY region""",
        )
        self.assertRecipeCSV(
            r,
            """
            region,pop,region_id
            Northeast,609480,Northeast
            South,5685230,South
            """,
        )

    def test_recipe_as_subquery(self):
        """Use a recipe subquery as a source for generating a new shelf."""

        recipe = (
            Recipe(shelf=self.census_shelf, session=self.session)
            .metrics("pop2000")
            .dimensions("state")
        )

        # Another approach is to use
        subq = recipe.subquery()
        recipe_subquery_shelf = Shelf({"pop": Metric(func.avg(subq.c.pop2000))})
        r2 = Recipe(shelf=recipe_subquery_shelf, session=self.session).metrics("pop")
        self.assertRecipeSQL(
            r2,
            """SELECT avg(anon_1.pop2000) AS pop
FROM
  (SELECT census.state AS state,
          sum(census.pop2000) AS pop2000
   FROM census
   GROUP BY state) AS anon_1""",
        )
        self.assertRecipeCSV(r2, "pop\n3147355.0")


class CacheContextTestCase(RecipeTestCase):
    def test_cache_context(self):
        recipe = self.recipe().metrics("age").dimensions("last")
        recipe.cache_context = "foo"

        self.assertEqual(len(recipe.all()), 2)
        self.assertEqual(recipe._cauldron["last"].cache_context, "foo")



from datetime import date, datetime
from time import gmtime

import dateparser
from sqlalchemy import Boolean, Date, DateTime, Integer, Numeric, String
from sqlalchemy.exc import CompileError


def convert_date(v):
    """Convert a passed parameter to a date if possible"""
    if v is None:
        return v
    elif isinstance(v, str):
        try:
            dt = dateparser.parse(v)
            return dt.date() if dt is not None else v
        except ValueError:
            return v
    elif isinstance(v, (float, int)):
        # Convert to a date
        tm = gmtime(v)
        return date(tm.tm_year, tm.tm_mon, tm.tm_mday)
    elif isinstance(v, dict) and "value" in v:
        v["value"] = convert_date(v["value"])
        return v
    else:
        return v


def convert_datetime(v):
    """Convert a passed parameter to a datetime if possible"""
    if v is None:
        return v
    elif isinstance(v, str):
        try:
            dt = dateparser.parse(v)
            return dt if dt is not None else v
        except ValueError:
            return v
    elif isinstance(v, (float, int)):
        # Convert to a date
        return datetime.utcfromtimestamp(v)
    elif isinstance(v, dict) and "value" in v:
        v["value"] = convert_datetime(v["value"])
        return v
    else:
        return v


def datatype_from_column_expression(c):
    """Determine a datatype from a column or column expression"""
    datatype = "unknown"
    try:
        if hasattr(c, "type") and c.type:
            # Check supported column types
            if isinstance(c.type, String):
                datatype = "str"
            elif isinstance(c.type, Date):
                datatype = "date"
            elif isinstance(c.type, DateTime):
                datatype = "datetime"
            elif isinstance(c.type, (Integer, Numeric)):
                datatype = "num"
            elif isinstance(c.type, Boolean):
                datatype = "bool"
    except CompileError:
        pass

    if datatype == "unknown":
        col_expr = str(c).lower()
        if col_expr.startswith("date"):
            datatype = "date"
        elif col_expr.startswith("timestamp"):
            datatype = "datetime"
    return datatype


def determine_datatype(ingr, role="value"):
    """Determine the datatype of an ingredients role

    Developers note: For ingredients constructed from parsed expressions
    this will use the datatype determined by the parser. This will be accurate.
    The fallback for ingredients constructed from raw SQLAlchemy will
    be to examine the column.type or compile the expression and try to
    make inference. This is less accurate.
    """
    from recipe.ingredients import Dimension, Filter, Having, Metric

    if isinstance(ingr, Dimension) and ingr.datatype_by_role:
        return ingr.datatype_by_role.get(role, None)
    elif isinstance(ingr, (Filter, Having)):
        return "bool"
    elif isinstance(ingr, (Metric)):
        return ingr.datatype



# -*- coding: utf-8 -*-
import os

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]
if os.getenv("SPELLCHECK"):
    extensions += ("sphinxcontrib.spelling",)
    spelling_show_suggestions = True
    spelling_lang = "en_US"

source_suffix = ".rst"
master_doc = "index"
project = "Recipe"
year = "2019"
author = "Chris Gemignani"
copyright = "{0}, {1}".format(year, author)
version = release = "0.36.4"

pygments_style = "trac"
templates_path = ["."]
extlinks = {
    "issue": ("https://github.com/juiceinc/recipe/issues/%s", "#"),
    "pr": ("https://github.com/juiceinc/recipe/pull/%s", "PR #"),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if not on_rtd:  # only set the theme if we're building docs locally
    html_theme = "sphinx_rtd_theme"

html_theme_options = {
    # 'canonical_url': '',
    # 'analytics_id': 'UA-XXXXXXX-1',  #  Provided by Google in your dashboard
    "logo_only": False,
    "display_version": True,
    "prev_next_buttons_location": "bottom",
    "style_external_links": False,
    # 'vcs_pageview_mode': '',
    "style_nav_header_background": "#4eb2cd",
    # Toc options
    "collapse_navigation": True,
    "sticky_navigation": True,
    "navigation_depth": 4,
    "includehidden": True,
    "titles_only": False,
}

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
}
html_short_title = "%s-%s" % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



import re
from datetime import date, datetime
from typing import List, Optional

import attr
import structlog
from sqlalchemy import Boolean, Date, DateTime, Float, Integer, String, alias, cast
from sqlalchemy.ext.declarative.api import DeclarativeMeta
from sqlalchemy.sql.base import ColumnCollection
from sqlalchemy.sql.sqltypes import Numeric

SLOG = structlog.get_logger(__name__)


VALID_COLUMN_RE = re.compile("^\w[\w ]*$")


def is_valid_column(colname: str) -> bool:
    """We can only build columns on field names that are alphanumeric"""
    return bool(VALID_COLUMN_RE.match(colname))


@attr.s
class Col:
    """Link a sqlalchemy column with a grammar rule"""

    datatype: str = attr.ib()
    sqla_col = attr.ib()
    name: str = attr.ib()
    idx: int = attr.ib(default=-1)
    namespace: str = attr.ib(default="")

    @classmethod
    def make_from_sqla_col(cls, sqla_col):
        if is_valid_column(sqla_col.name):
            if isinstance(sqla_col.type, String):
                datatype = "str"
            elif isinstance(sqla_col.type, Date):
                datatype = "date"
            elif isinstance(sqla_col.type, DateTime):
                datatype = "datetime"
            elif isinstance(sqla_col.type, Integer):
                datatype = "num"
            elif isinstance(sqla_col.type, Numeric):
                datatype = "num"
            elif isinstance(sqla_col.type, Boolean):
                datatype = "bool"
            elif str(sqla_col.type).startswith("TIMESTAMP"):
                # This check is to support snowflake which uses custom
                # types for timestamps that are not derived from sqlalchemy
                # types.
                datatype = "datetime"
            else:
                datatype = "unusable"
            return cls(
                namespace="", datatype=datatype, sqla_col=sqla_col, name=sqla_col.name
            )

    @classmethod
    def make_from_constant(cls, key, value):
        """Make a column from a scalar constant"""
        if is_valid_column(key):
            datatype = "unusable"
            sqla_col = None

            if isinstance(value, str):
                datatype = "str"
                sqla_col = cast(value, String)
            elif isinstance(value, date):
                datatype = "date"
                sqla_col = cast(value, Date)
            elif isinstance(value, datetime):
                datatype = "datetime"
                sqla_col = cast(value, DateTime)
            elif isinstance(value, int):
                datatype = "num"
                sqla_col = cast(value, Integer)
            elif isinstance(value, float):
                datatype = "num"
                sqla_col = cast(value, Float)
            elif isinstance(value, bool):
                datatype = "bool"
                sqla_col = cast(value, Boolean)
            return cls(namespace="", datatype=datatype, sqla_col=sqla_col, name=key)

    @property
    def rule_name(self) -> str:  # sourcery skip: raise-specific-error
        """The name for the lark rule for this column"""
        if self.idx == -1:
            raise Exception("Must assign indexes first")
        return f"{self.datatype}_{self.idx}"

    @property
    def field_name(self) -> str:
        """What to call this column in expressions."""
        return f"{self.namespace}\\.{self.name}" if self.namespace else self.name

    def as_rule(self):
        if " " in self.field_name:
            return f'    {self.rule_name}: "[" + /{self.field_name}/i + "]"'
        else:
            return f'    {self.rule_name}: "[" + /{self.field_name}/i + "]" | /{self.field_name}/i'


@attr.s
class ColCollection:
    """A collection of columns. These columns may come from more than one selectable."""

    columns: List[Col] = attr.ib()

    def assign_indexes(self):
        """Sort columns by datatype and assign an index"""
        idx = 0
        prev_datatype = None
        for col in sorted(self.columns, key=lambda c: (c.datatype, c.name)):
            if col.datatype != prev_datatype:
                idx = 0
            col.idx = idx
            prev_datatype = col.datatype
            idx += 1

    def set_namespace(self, namespace):
        for c in self.columns:
            c.namespace = namespace

    def extend(self, other_cc):
        self.columns += other_cc.columns

    def column_lookup(self) -> dict:
        """Generate a lookup from rule names to the sqlalchemy columns"""
        return {c.rule_name: c.sqla_col for c in self.columns}


def make_column_collection_for_selectable(
    selectable, *, namespace: Optional[str] = None
) -> ColCollection:
    """Return a dictionary of columns. The keys
    are unique lark rule names prefixed by the column type
    like num_0, num_1, string_0, etc.

    The values are the selectable column reference
    """
    from recipe import Recipe

    if isinstance(selectable, Recipe):
        selectable = selectable.subquery()

    if isinstance(selectable, DeclarativeMeta):
        column_iterable = selectable.__table__.columns
    # Selectable is a sqlalchemy subquery
    elif hasattr(selectable, "c") and isinstance(selectable.c, ColumnCollection):
        column_iterable = selectable.c
    else:
        raise Exception(
            f"Selectable must be a Table or ColumnCollection, this is a {type(selectable)}"
        )

    columns = []
    for sqla_col in column_iterable:
        if col := Col.make_from_sqla_col(sqla_col):
            columns.append(col)

    cc = ColCollection(columns)
    if namespace:
        cc.set_namespace(namespace=namespace)
    return cc


def is_constant_expression(v) -> bool:
    return isinstance(v, str) and "(" in v and ")" in v


def has_constant_expressions(constants: dict) -> bool:
    return any(is_constant_expression(v) for v in constants.values())


def has_constant_literals(constants: dict) -> bool:
    return any(not is_constant_expression(v) for v in constants.values())


def make_column_collection_for_constant_literals(
    constants: dict, *, namespace: Optional[str] = None
) -> ColCollection:
    """
    Constants are a dict of names to scalar values to use in expressions
    We will treat them as if they are another column collection

    Args:
        selectable: A selectable for column expressions
        constants (dict): A dict with string keys and scalar values
        namespace (str, optional): A namespace to add. Defaults to None.

    Returns:
        ColumnCollection: A column collection of constant values
    """
    # Create columns
    constant_columns = [
        Col.make_from_constant(k, v)
        for k, v in constants.items()
        if not is_constant_expression(v)
    ]
    cc = ColCollection(constant_columns)
    if namespace:
        cc.set_namespace(namespace=namespace)
    return cc


def make_column_collection_for_constant_expressions(
    builder, constants: dict, *, namespace: Optional[str] = None
) -> ColCollection:
    """
    Constants are a dict of names to scalar values to use in expressions
    We will treat them as if they are another column collection

    Args:
        builder: A selectable for column expressions
        constants (dict): A dict with string keys and scalar values
        namespace (str, optional): A namespace to add. Defaults to None.

    Returns:
        ColumnCollection: A column collection of constant values
    """
    # Create columns
    from sqlalchemy import select

    expression_columns = []
    sel = select()
    for k, v in constants.items():
        if is_constant_expression(v):
            expr, dtype = builder.parse(v)
            expression_columns.append(expr.label(k))

    sel = alias(select(expression_columns), "constants")
    return make_column_collection_for_selectable(sel, namespace=namespace)


def make_columns_grammar(cc: ColumnCollection) -> str:
    """Return a lark rule that looks like

    // These are my raw columns
    str_0: "[" + /username/i + "]" | /username/i
    str_1: "[" + /department/i + "]" | /department/i
    str_2: "[" + /testid/i + "]" | /testid/i
    num_0: "[" + /score/i + "]" | /score/i
    """
    cc.assign_indexes()
    return "\n".join(sorted([col.as_rule() for col in cc.columns]))


def gather_columns(
    datatype_rule_name: str,
    cc: ColumnCollection,
    datatype: str,
    *,
    additional_rules=None,
) -> str:
    """Build a list of all column rules matching a datatype along with potential additional rules."""
    if additional_rules is None:
        additional_rules = []

    matching_cols = [c for c in cc.columns if c.datatype == datatype]
    column_rules = [f"{datatype}_{n}" for n in range(len(matching_cols))]
    if matching_cols + additional_rules:
        raw_rule_name = datatype_rule_name.split(".")[0]

        # Reduce a pair of parens around a type back to itself.
        paren_rule = f'"(" + {raw_rule_name} + ")"'

        return f"{datatype_rule_name}: " + " | ".join(
            column_rules + additional_rules + [paren_rule]
        )
    else:
        return f'{datatype_rule_name}: "DUMMYVALUNUSABLECOL"'


def make_grammar(columns):
    """Build a grammar for this selectable using columns"""
    grammar = f"""
    col: boolean | string | num | date | datetime_end | datetime | unusable_col | unknown_col | error_math | error_vector_expr | error_not_nonboolean | error_between_expr | error_aggr | error_if_statement
    //paren_col: "(" col ")" -> col

    // These are the raw columns in the selectable
    {make_columns_grammar(columns)}

    {gather_columns("unusable_col", columns, "unusable")}
    {gather_columns("date.1", columns, "date", additional_rules=["date_conv", "date_fn", "day_conv", "week_conv", "month_conv", "quarter_conv", "year_conv", "dt_day_conv", "dt_week_conv", "dt_month_conv", "dt_quarter_conv", "dt_year_conv", "datetime_to_date_conv", "date_aggr", "date_if_statement", "date_coalesce", "lastday"])}
    {gather_columns("datetime.2", columns, "datetime", additional_rules=["datetime_conv", "datetime_if_statement", "datetime_coalesce"])}
    // Datetimes that are converted to the end of day
    {gather_columns("datetime_end.1", columns, "datetime", additional_rules=["datetime_end_conv", "datetime_aggr"])}
    {gather_columns("boolean.1", columns, "bool", additional_rules=["TRUE", "FALSE", "bool_expr", "date_bool_expr", "datetime_bool_expr", "str_like_expr", "vector_expr", "between_expr", "date_between_expr", "datetime_between_expr", "not_boolean", "or_boolean", "and_boolean", "paren_boolean", "intelligent_date_expr", "intelligent_datetime_expr"])}
    {gather_columns("string.1", columns, "str", additional_rules=["ESCAPED_STRING", "string_add", "string_cast", "string_coalesce", "string_substr", "string_if_statement", "string_aggr"])}
    {gather_columns("num.1", columns, "num", additional_rules=["NUMBER", "num_add", "num_sub", "num_mul", "num_div", "int_cast", "num_coalesce", "aggr", "error_aggr", "num_if_statement", "age_conv", "datediff", "extract"])}
    string_add: string "+" string
    num_add.1: num "+" num | "(" num "+" num ")"
    num_sub.1: num "-" num | "(" num "-" num ")"
    num_mul.2: num "*" num | "(" num "*" num ")"
    num_div.2: num "/" num | "(" num "/" num ")"
    add: col "+" col | "(" col "+" col ")"

    // Various error conditions (fields we don't recognize, bad math)
    // Low priority matching of any [columnname] values
    unknown_col.0: "[" + NAME + "]" | NAME
    error_math.0: error_add | error_sub | error_mul | error_div
    error_add.0: col "+" col
    error_sub.0: col "-" col
    error_mul.0: col "*" col
    error_div.0: col "/" col
    error_between_expr.0: col BETWEEN col AND col
    error_vector_expr.0: col vector_comparator mixedarray
    error_not_nonboolean: NOT string | NOT num
    mixedarray.0: "(" [CONSTANT ("," CONSTANT)*] ","? ")"
    CONSTANT: ESCAPED_STRING | NUMBER

    // Boolean scalar expressions like 'a > b'
    paren_boolean.5: "(" boolean ")"
    not_boolean.4: NOT boolean
    and_boolean.3: boolean AND boolean
    or_boolean.2: boolean OR boolean
    bool_expr: col comparator col | col null_comparator NULL
    str_like_expr: string LIKE ESCAPED_STRING
    date_bool_expr.1: date comparator (date | string)
    datetime_bool_expr.2: datetime comparator (datetime | string)
    comparator: EQ | NE | LT | LTE | GT | GTE
    null_comparator: EQ | NE | IS | IS NOT
    LIKE: /i?like/i
    EQ: "="
    NE: "!=" | "<>"
    LT: "<"
    LTE: "<="
    GT: ">"
    GTE: ">="

    // Boolean vector expressions like 'a in (array of constants)'
    intelligent_date_expr.1: date IS INTELLIGENT_DATE_OFFSET INTELLIGENT_DATE_UNITS
    intelligent_datetime_expr.1: datetime IS INTELLIGENT_DATE_OFFSET INTELLIGENT_DATE_UNITS
    between_expr.1: string BETWEEN string AND string | num BETWEEN num AND num | date BETWEEN date AND date | datetime BETWEEN datetime AND datetime_end
    date_between_expr.2: date BETWEEN (date | string) AND (date | string)
    datetime_between_expr.3: datetime BETWEEN (datetime | string) AND (datetime_end | string)

    vector_expr.1: string vector_comparator stringarray | num vector_comparator numarray
    vector_comparator.1: NOT? IN
    stringarray.1: "(" [ESCAPED_STRING ("," ESCAPED_STRING)*] ","? ")"  -> consistent_array
    numarray.1: "(" [NUMBER ("," NUMBER)*] ","?  ")"                    -> consistent_array

    // Date
    date_conv.3: /date/i "(" ESCAPED_STRING ")"
    date_fn.3: /date/i "(" num "," num "," num ")"
    datetime_to_date_conv.3: /date/i "(" datetime ")"  -> dt_day_conv
    datetime_conv.2: /date/i "(" ESCAPED_STRING ")"
    datetime_end_conv.1: /date/i "(" ESCAPED_STRING ")"

    // Conversions
    // date->date
    day_conv: /day/i "(" date ")"
    week_conv: /week/i "(" date ")"
    month_conv: /month/i "(" date ")"
    quarter_conv: /quarter/i "(" date ")"
    year_conv: /year/i "(" date ")"
    // datetime->date
    dt_day_conv: /day/i "(" datetime ")"
    dt_week_conv: /week/i "(" datetime ")"
    dt_month_conv: /month/i "(" datetime ")"
    dt_quarter_conv: /quarter/i "(" datetime ")"
    dt_year_conv: /year/i "(" datetime ")"
    /// date->int
    datediff: /datediff/i "(" date "," date ("," DATEPART )? ")"
    extract: /extract/i "(" DATEPART "," (date | datetime) ")"
    lastday: /lastday/i "(" (date | datetime) ("," DATEPART )? ")"
    // col->string
    string_cast: /string/i "(" col ")"
    string_substr: /substr/i "(" string "," [num ("," num)?] ")"
    // col->int
    int_cast: /int/i "(" col ")"
    // date->int
    age_conv: /age/i "(" (date | datetime) ")"
    // date->int
    // TODO: age_conv: /age/i "(" (date | datetime) ")"
    // TODO: date - date => int
    num_coalesce: /coalesce/i "(" num "," num ")"                 -> coalesce
    string_coalesce: /coalesce/i "(" string "," string ")"        -> coalesce
    date_coalesce: /coalesce/i "(" date "," date ")"              -> coalesce
    datetime_coalesce: /coalesce/i "(" datetime "," datetime ")"  -> coalesce

    // Aggregations that are errors
    error_aggr.0: error_sum_aggr | error_min_aggr | error_max_aggr | error_avg_aggr | error_median_aggr | error_percentile_aggr
    error_sum_aggr.0: /sum/i "(" col ")"
    error_min_aggr.0: /min/i "(" col ")"
    error_max_aggr.0: /max/i "(" col ")"
    error_avg_aggr.0: /avg/i "(" col ")" | /average/i "(" col ")"
    error_median_aggr.0: /median/i "(" col ")"
    error_percentile_aggr.0: /percentile\d\d?/i "(" col ")"
    // Aggregations that return numbers
    aggr.1: sum_aggr | min_aggr | max_aggr | avg_aggr | count_aggr | count_distinct_aggr | median_aggr | percentile_aggr
    sum_aggr.1: /sum/i "(" num ")"
    min_aggr.1: /min/i "(" num ")"
    max_aggr.1: /max/i "(" num ")"
    avg_aggr.1: /avg/i "(" num ")" | /average/i "(" num ")"
    count_aggr.1: /count/i "(" (num | string | date | datetime | boolean | star) ")"
    count_distinct_aggr.1: /count_distinct/i "(" (num | string | date | datetime | boolean) ")"
    median_aggr.1: /median/i "(" num ")"
    percentile_aggr.1: /percentile\d\d?/i "(" num ")"
    // Aggregations that return strings
    string_aggr.1: min_string_aggr | max_string_aggr
    min_string_aggr.1: /min/i "(" string ")"            -> min_aggr
    max_string_aggr.1: /max/i "(" string ")"            -> max_aggr
    // Aggregations that return dates
    date_aggr.1: min_date_aggr | max_date_aggr
    min_date_aggr.1: /min/i "(" date ")"            -> min_aggr
    max_date_aggr.1: /max/i "(" date ")"            -> max_aggr
    // Aggregations that return datetimes
    datetime_aggr.1: min_datetime_aggr | max_datetime_aggr
    min_datetime_aggr.1: /min/i "(" datetime ")"    -> min_aggr
    max_datetime_aggr.1: /max/i "(" datetime ")"    -> max_aggr

    // functions
    num_if_statement.4: IF "(" (boolean "," (num | NULL) ","?)+ (num | NULL)? ")"                    -> if_statement
    string_if_statement.4: IF "(" (boolean "," (string | NULL) ","?)+ (string | NULL)? ")"           -> if_statement
    date_if_statement.4: IF "(" (boolean "," (date | NULL) ","?)+ (date | NULL)? ")"                 -> if_statement
    datetime_if_statement.4: IF "(" (boolean "," (datetime | NULL) ","?)+ (datetime | NULL)? ")"     -> if_statement
    //error_if_statement.3: IF "(" (col "," (col | NULL) ","?)+ (col | NULL)? ")"
    error_if_statement.3: IF "(" (col "," (col | NULL) ","?)+ (col | NULL)? ")"

    star: "*"
    TRUE: /TRUE/i
    FALSE: /FALSE/i
    OR: /OR/i
    AND: /AND/i
    NOT: /NOT/i
    IN: /IN/i
    IS: /IS/i
    BETWEEN: /BETWEEN/i
    NULL: /NULL/i
    IF: /IF/i
    INTELLIGENT_DATE_OFFSET: /prior/i | /last/i | /previous/i | /current/i | /this/i | /next/i
    INTELLIGENT_DATE_UNITS: /ytd/i | /year/i | /qtr/i | /month/i | /mtd/i | /day/i
    DATEPART: /dayofweek/i | /day/i | /dayofyear/i | /week/i | /week/i "(" WEEKDAY ")" | /month/i | /quarter/i | /year/i | /isoyear/i | /isoweek/i
    WEEKDAY: /monday/i | /tuesday/i | /wednesday/i | /thursday/i | /friday/i | /saturday/i | /sunday/i
    COMMENT: /#.*/

    %import common.CNAME                       -> NAME
    %import common.SIGNED_NUMBER               -> NUMBER
    %import common.ESCAPED_STRING
    %import common.WS_INLINE
    %ignore COMMENT
    %ignore WS_INLINE
"""
    return grammar



"""Shelf config _version="2" supports parsed fields using a lark parser."""
from copy import deepcopy
from lark.exceptions import LarkError
import logging
from sureberus import schema as S

from .utils import (
    coerce_format,
    coerce_pop_version,
    coerce_shelf_meta,
    _chain,
    SCALAR_TYPES,
)

logging.captureWarnings(True)


def move_extra_fields(value):
    """Move any fields that look like "{role}_field" into the extra_fields
    list. These will be processed as fields. Rename them as {role}_expression.
    """
    if isinstance(value, dict):
        keys_to_move = [k for k in value.keys() if k.endswith("_field")]
        if keys_to_move:
            value["extra_fields"] = []
            for k in sorted(keys_to_move):
                value["extra_fields"].append(
                    {"name": k[:-6] + "_expression", "field": value.pop(k)}
                )

    return value


def coerce_replace_refs(shelf):
    """Replace a reference in a field like @foo with the contents of foo's field"""
    replacements = []
    for k, v in shelf.items():
        if "field" in v:
            replacements.append(("@" + k, v["field"]))

    # Sort in descending order of length then by k
    replacements.sort(key=lambda k: (-1 * len(k[0]), k[0]))

    # Search for fields and replace with their replacements
    for ingr in shelf.values():
        for k in ingr.keys():
            if k == "field" or k.endswith("_field"):
                v = ingr[k]
                if isinstance(v, str) and "@" in v:
                    for search_for, replace_with in replacements:
                        v = v.replace(search_for, replace_with)
                    ingr[k] = v

    return shelf


def _stringify(value):
    """Convert a value into a string that will parse"""
    if value is None:
        return "NULL"
    elif value is True:
        return "TRUE"
    elif value is False:
        return "FALSE"
    elif isinstance(value, str):
        return '"' + value.replace('"', '\\"') + '"'
    else:
        return str(value)


def _convert_bucket_to_field(bucket, bucket_default_label, use_indices=False):
    """Convert a bucket structure to a if function expression field.

    This assumes that all conditions have been converted from
    partial conditions like '<5' to full conditions like 'age<5'
    with _convert_partial_conditions
    """
    parts, idx = [], 0
    for itm in bucket:
        cond = itm.get("condition")
        label = itm.get("label")
        parts.append(cond)
        if use_indices:
            parts.append(str(idx))
        else:
            # Stringify the label
            parts.append(_stringify(label))
        idx += 1

    # Add the default value
    if bucket_default_label is None:
        bucket_default_label = "Not found"
    if use_indices:
        parts.append(str(9999))
    else:
        parts.append(_stringify(bucket_default_label))

    return "if(" + ",".join(parts) + ")"


def _lowercase_kind(value):
    """Ensure kind is lowercase with a default of "metric" """
    if isinstance(value, dict):
        kind = value.get("kind", "metric").lower()
        # measure is a synonym for metric
        if kind == "measure":
            kind = "metric"
        value["kind"] = kind

    return value


def _save_raw_config(value):
    """Save the original config excluding _config and _meta"""
    config = deepcopy(value)
    config.pop("_config", None)
    config.pop("_neta", None)
    value["_config"] = config
    return value


# A field that may OR MAY NOT contain an aggregation.
# It will be the transformers responsibility to add an aggregation if one is missing
field_schema = S.String(required=True)

labeled_condition_schema = S.Dict(
    schema={"condition": field_schema, "label": S.String(required=True)}
)

# A full condition guaranteed to not contain an aggregation
named_condition_schema = S.Dict(
    schema={"condition": field_schema, "name": S.String(required=True)}
)

format_schema = S.String(coerce=coerce_format, required=False)


metric_schema = S.Dict(
    schema={
        "field": field_schema,
        "format": format_schema,
        "quickselects": S.List(required=False, schema=labeled_condition_schema),
    },
    allow_unknown=True,
)

dimension_schema = S.Dict(
    schema={
        "field": field_schema,
        "extra_fields": S.List(
            required=False,
            schema=S.Dict(
                schema={"field": field_schema, "name": S.String(required=True)}
            ),
        ),
        "buckets": S.List(required=False, schema=labeled_condition_schema),
        "buckets_default_label": {"anyof": SCALAR_TYPES, "required": False},
        "format": format_schema,
        "lookup": S.Dict(required=False),
        "quickselects": S.List(required=False, schema=named_condition_schema),
    },
    coerce=move_extra_fields,
    allow_unknown=True,
)

filter_schema = S.Dict(allow_unknown=True, schema={"condition": field_schema})

having_schema = S.Dict(allow_unknown=True, schema={"condition": field_schema})

ingredient_schema = S.Dict(
    choose_schema=S.when_key_is(
        "kind",
        {
            "metric": metric_schema,
            "dimension": dimension_schema,
            "filter": filter_schema,
            "having": having_schema,
        },
        default_choice="metric",
    ),
    coerce=_chain(_lowercase_kind, _save_raw_config),
    registry={},
)


strict_metric_schema = S.Dict(
    allow_unknown=False,
    schema={
        # _version is deprecated.
        "_version": S.String(required=False),
        "_shelf_meta": S.Dict(required=False),
        "icon": S.String(required=False),
        "_meta": S.Dict(required=False),
        "_config": S.Dict(required=False),
        "filter": S.String(required=False),
        "singular": S.String(required=False),
        "plural": S.String(required=False),
        "field": field_schema,
        "format": format_schema,
        "quickselects": S.List(required=False, schema=labeled_condition_schema),
    },
)

strict_dimension_schema = S.Dict(
    allow_unknown=False,
    schema={
        # _version is deprecated.
        "_version": S.String(required=False),
        "_shelf_meta": S.Dict(required=False),
        "icon": S.String(required=False),
        "_meta": S.Dict(required=False),
        "_config": S.Dict(required=False),
        "filter": S.String(required=False),
        "date_aggregation": S.String(required=False, allowed=["year", "month", "day"]),
        "singular": S.String(required=False),
        "plural": S.String(required=False),
        "field": field_schema,
        "extra_fields": S.List(
            required=False,
            schema=S.Dict(
                schema={"field": field_schema, "name": S.String(required=True)}
            ),
        ),
        "buckets": S.List(required=False, schema=labeled_condition_schema),
        "buckets_default_label": {"anyof": SCALAR_TYPES, "required": False},
        "format": format_schema,
        "lookup": S.Dict(required=False),
        "quickselects": S.List(required=False, schema=named_condition_schema),
    },
    coerce=move_extra_fields,
)

strict_filter_schema = S.Dict(
    allow_unknown=False,
    schema={
        # This _version is deprecated.
        "_version": S.String(required=False),
        "_shelf_meta": S.Dict(required=False),
        "condition": field_schema,
    },
)

strict_having_schema = S.Dict(
    allow_unknown=False,
    schema={
        # _version is deprecated.
        "_version": S.String(required=False),
        "_shelf_meta": S.Dict(required=False),
        "condition": field_schema,
    },
)


# A schema for ingredients that only allows specified fields
strict_ingredient_schema = S.Dict(
    choose_schema=S.when_key_is(
        "kind",
        {
            "metric": strict_metric_schema,
            "dimension": strict_dimension_schema,
            "filter": strict_filter_schema,
            "having": strict_having_schema,
        },
        default_choice="metric",
    ),
    coerce=_chain(_lowercase_kind, _save_raw_config),
    registry={},
)


# A schema for a shelf with fields turned into sqlalchemy using lark parsing
# and allow_unknown properties on the definitions
shelf_schema = S.Dict(
    valueschema=ingredient_schema,
    keyschema=S.String(),
    coerce=_chain(coerce_pop_version, coerce_shelf_meta, coerce_replace_refs),
    allow_unknown=True,
)

# A schema for a shelf with fields turned into sqlalchemy using lark parsing
# and NO allow_unknown on the ingredient definitions
strict_shelf_schema = S.Dict(
    valueschema=strict_ingredient_schema,
    keyschema=S.String(),
    coerce=_chain(coerce_pop_version, coerce_shelf_meta, coerce_replace_refs),
    allow_unknown=True,
)



from sureberus import schema as S
from .parsed_schemas import shelf_schema

# This schema is used with sureberus
recipe_schema = S.Dict(
    schema={
        "metrics": S.List(schema=S.String(), required=False),
        "dimensions": S.List(schema=S.String(), required=False),
        "filters": S.List(schema=S.String(), required=False),
        "order_by": S.List(schema=S.String(), required=False),
    }
)



import math
import re
import unicodedata
from recipe.schemas import recipe_schema
from sureberus import schema as S

from sqlalchemy.sql.functions import FunctionElement

WHITESPACE_RE = re.compile(r"\s+", flags=re.DOTALL | re.MULTILINE)


def replace_whitespace_with_space(s):
    """Replace multiple whitespaces with a single space."""
    return WHITESPACE_RE.sub(" ", s)


def clean_unicode(value):
    """Convert value into ASCII bytes by brute force."""
    if not isinstance(value, str):
        value = str(value)
    try:
        return value.encode("ascii")
    except UnicodeEncodeError:
        value = unicodedata.normalize("NFKD", value)
        return value.encode("ascii", "ignore")


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


def disaggregate(expr):
    if isinstance(expr, FunctionElement):
        return expr.clause_expr
    else:
        return expr


def pad_values(values, prefix="RECIPE-DUMMY-VAL-", bin_size=11):
    """Redshift recompiles queries when a where IN clause includes a
    different number of values. To avoid this, pad out the list
    of string values to certain fixed lengths.

    The default bin size is 11. During testing we discovered that
    compilation was required for where clauses with different numbers
    of items but all queries with 11 or more items shared a compiled
    query.
    """
    assert isinstance(values, (list, tuple))
    cnt = len(values)
    if cnt and isinstance(values[0], str):
        # Round up to the nearest bin_size
        desired_cnt = int(math.ceil(float(cnt) / bin_size) * bin_size)
        added_values = [prefix + str(i + 1) for i in range(desired_cnt - cnt)]
        if isinstance(values, tuple):
            return values + tuple(added_values)
        else:
            return values + added_values
    else:
        return values


def make_schema(recipe_extensions: list) -> dict:
    """Make a sureberus schema to validate a recipe.from_config."""
    schema = recipe_schema["schema"].copy()
    for ext in recipe_extensions:
        additional_schema = getattr(ext, "recipe_schema", {})
        schema.update(additional_schema)
    return S.Dict(schema=schema)



import attr
import structlog
from lark import Tree, Visitor
from lark.lexer import Token

SLOG = structlog.get_logger(__name__)


@attr.s
class SQLALchemyValidator(Visitor):
    """
    Visit the tree and return descriptive information. Populate
    a list of errors.

    Args:
        text (str): A copy of the parsed text for error descriptions
        forbid_aggregation (bool): Should aggregations be treated as an error
        drivername (str): The database engine we are running against
    """

    text = attr.ib()
    forbid_aggregation = attr.ib()
    drivername = attr.ib()
    # Was an aggregation encountered in the tree?
    found_aggregation = attr.ib(default=False)
    # What is the datatype of the returned expression
    last_datatype = attr.ib(default=None)
    # Errors encountered while visiting the tree
    errors = attr.ib(factory=list)

    def _data_type(self, tree):
        # Find the data type for a tree
        if tree is None:
            return None
        if tree.data == "col":
            dt = self._data_type(tree.children[0])
        else:
            dt = tree.data
        dt = str(dt)  # Convert Tokens to strings
        if dt == "datetime_end":
            dt = "datetime"
        elif dt == "string":
            dt = "str"
        elif dt == "boolean":
            dt = "bool"
        return dt

    def _add_error(self, message, tree):
        """Add an error pointing to this location in the parsed string"""
        tok = None
        # Find the first token
        while tree and tree.children:
            tree = tree.children[0]
            if isinstance(tree, Token):
                tok = tree
                break

        if tok:
            extra_context = self._get_context_for_token(tok, span=200)
            message = f"{message}\n\n{extra_context}"
        self.errors.append(message)

    def _get_context_for_token(self, tok, span=40):
        pos = tok.start_pos
        start = max(pos - span, 0)
        end = pos + span
        before = self.text[start:pos].rsplit("\n", 1)[-1]
        after = self.text[pos:end].split("\n", 1)[0]
        return before + after + "\n" + " " * len(before) + "^\n"

    def _error_math(self, tree, verb):
        tok1 = tree.children[0].children[0]
        tok2 = tree.children[1].children[0]
        self._add_error(f"{tok1.data} and {tok2.data} can not be {verb}", tree)

    def col(self, tree):
        self.last_datatype = self._data_type(tree)

    def error_add(self, tree):
        self._error_math(tree, "added together")

    def error_mul(self, tree):
        self._error_math(tree, "multiplied together")

    def error_sub(self, tree):
        self._error_math(tree, "subtracted")

    def error_div(self, tree):
        self._error_math(tree, "divided")

    def error_if_statement(self, tree):
        args = tree.children
        # Throw away the "if"
        args = args[1:]

        # If there's an odd number of args, pop the last one to use as the else
        if len(args) % 2:
            else_expr = args.pop()
        else:
            else_expr = None

        # The "odd" args should be booleans
        bool_args = args[::2]
        # The "even" args should be values of the same type
        value_args = args[1::2]

        # Check that the boolean args are boolean
        for arg in bool_args:
            dt = self._data_type(arg)
            if dt != "bool":
                self._add_error("This should be a boolean column or expression", arg)

        # Data types of columns must match
        value_type = None
        for arg in value_args + [else_expr]:
            dt = self._data_type(arg)
            if dt is not None:
                if value_type is None:
                    value_type = dt
                elif value_type != dt:
                    self._add_error(
                        f"The values in this if statement must be the same type, not {value_type} and {dt}",
                        arg,
                    )

    def aggr(self, tree):
        self.found_aggregation = True
        if self.forbid_aggregation:
            self._add_error("Aggregations are not allowed in this field.", tree)

    def string_aggr(self, tree):
        self.found_aggregation = True
        if self.forbid_aggregation:
            self._add_error("Aggregations are not allowed in this field.", tree)

    def date_aggr(self, tree):
        self.found_aggregation = True
        if self.forbid_aggregation:
            self._add_error("Aggregations are not allowed in this field.", tree)

    def datetime_aggr(self, tree):
        self.found_aggregation = True
        if self.forbid_aggregation:
            self._add_error("Aggregations are not allowed in this field.", tree)

    def unknown_col(self, tree):
        """Column name doesn't exist in the data"""
        tok1 = tree.children[0]
        self._add_error(f"{tok1} is not a valid column name", tree)

    def unusable_col(self, tree):
        """Column name isn't a data type we can handle"""
        tok1 = tree.children[0]
        self._add_error(
            f"{tok1} is a data type that can't be used. Usable data types are strings, numbers, boolean, dates, and datetimes",
            tree,
        )

    def error_not_nonboolean(self, tree):
        """NOT string or NOT num"""
        self._add_error("NOT requires a boolean value", tree)

    def mixedarray(self, tree):
        """An array containing a mix of strings and numbers"""
        self._add_error("An array may not contain both strings and numbers", tree)

    def vector_expr(self, tree):
        val, comp, arr = tree.children
        # If the left hand side is a number or string primitive
        if isinstance(val.children[0], Token) and val.children[0].type in (
            "NUMBER",
            "ESCAPED_STRING",
        ):
            self._add_error("Must be a column or expression", val)

    def error_aggr(self, tree):
        """Aggregating a bad data type"""
        fn = tree.children[0].children[0]
        dt = self._data_type(tree.children[0].children[1])
        self._add_error(f"A {dt} can not be aggregated using {fn}.", tree)

    def error_between_expr(self, tree):
        col, BETWEEN, left, AND, right = tree.children
        col_type = self._data_type(col)
        left_type = self._data_type(left)
        right_type = self._data_type(right)
        if col_type == "datetime":
            if left_type == "date":
                left_type = "datetime"
            if right_type == "date":
                right_type = "datetime"
        if not (col_type == left_type == right_type):
            self._add_error(
                f"When using between, the column ({col_type}) and between values ({left_type}, {right_type}) must be the same data type.",
                tree,
            )

    def bool_expr(self, tree):
        """a > b where the types of a and b don't match"""
        left, _, right = tree.children
        if isinstance(left, Tree) and isinstance(right, Tree):
            left_data_type = self._data_type(left)
            right_data_type = self._data_type(right)
            if left_data_type == right_data_type == "date":
                return
            if left_data_type == right_data_type == "datetime":
                return
            if left_data_type in ("date", "datetime") and right_data_type == "string":
                # Strings will be auto converted
                return
            if left_data_type != right_data_type:
                self._add_error(
                    f"Can't compare {left_data_type} to {right_data_type}", tree
                )

    def percentile_aggr(self, tree):
        """Sum up the things"""
        percentile, fld = tree.children
        percentile_val = int(percentile[len("percentile") :])
        if percentile_val not in (1, 5, 10, 25, 50, 75, 90, 95, 99):
            self._add_error(
                f"Percentile values of {percentile_val} are not supported.", tree
            )
        if self.drivername == "sqlite":
            self._add_error("Percentile is not supported on sqlite", tree)



from unittest.mock import patch
from recipe.dynamic_extensions import DynamicExtensionBase, run_hooks


class ToyExtension(DynamicExtensionBase):
    def execute(self):
        return super(ToyExtension, self).execute()


class ToyExtension2(DynamicExtensionBase):
    name = "TestExtension2"

    def execute(self):
        self.recipe_parts["test"] = 2
        return self.recipe_parts


def test_use_dyanmic_extension_base():
    recipe_parts = {"test": "value"}
    ext = ToyExtension(recipe_parts)
    assert ext.hook_type == "modify_query"
    assert ext.recipe_parts == recipe_parts
    assert ext.execute() == recipe_parts


@patch("recipe.dynamic_extensions.NamedExtensionManager")
def test_run_hooks_no_extensions(nem_patch):
    recipe_parts = {"test": 1}
    assert recipe_parts == run_hooks(recipe_parts, "modify_query")
    assert not nem_patch.called


def test_run_hooks():
    recipe_parts = {"test": 1}
    expected_result = {"test": 2}
    result = run_hooks(recipe_parts, hook_type="testing", extensions=["toyextension2"])
    assert expected_result == result



# -*- coding: utf-8 -*-
from datetime import date

import pytest
from faker import Faker
from faker.providers import BaseProvider
from tests.test_base import RecipeTestCase
from recipe.utils import (
    AttrDict,
    FakerAnonymizer,
    FakerFormatter,
    replace_whitespace_with_space,
    generate_faker_seed,
    pad_values,
    make_schema,
)

uppercase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"


class TestUtils(object):
    def test_replace_whitespace_with_space(self):
        assert replace_whitespace_with_space("fooo    moo") == "fooo moo"
        assert replace_whitespace_with_space("fooo\n\t moo") == "fooo moo"

    def test_generate_faker_seed(self):
        assert generate_faker_seed(None) == 7701040980221191251
        assert generate_faker_seed(0) == 14973660089898329583
        assert generate_faker_seed("hi") == 5329599339005471788
        assert generate_faker_seed([]) == 15515306683186839187


class PadValuesTestCase(RecipeTestCase):
    def test_pad_values(self):
        """A list or tuple of values are padded to a multiple of bin size"""
        with self.assertRaises(Exception):
            pad_values("rocket")

        v = pad_values([])
        assert v == []

        v = pad_values(tuple())
        assert v == tuple()

        v = pad_values(("hi",))
        assert (
            "hi",
            "RECIPE-DUMMY-VAL-1",
            "RECIPE-DUMMY-VAL-2",
            "RECIPE-DUMMY-VAL-3",
            "RECIPE-DUMMY-VAL-4",
            "RECIPE-DUMMY-VAL-5",
            "RECIPE-DUMMY-VAL-6",
            "RECIPE-DUMMY-VAL-7",
            "RECIPE-DUMMY-VAL-8",
            "RECIPE-DUMMY-VAL-9",
            "RECIPE-DUMMY-VAL-10",
        ) == v
        v = pad_values(["hi"])
        assert [
            "hi",
            "RECIPE-DUMMY-VAL-1",
            "RECIPE-DUMMY-VAL-2",
            "RECIPE-DUMMY-VAL-3",
            "RECIPE-DUMMY-VAL-4",
            "RECIPE-DUMMY-VAL-5",
            "RECIPE-DUMMY-VAL-6",
            "RECIPE-DUMMY-VAL-7",
            "RECIPE-DUMMY-VAL-8",
            "RECIPE-DUMMY-VAL-9",
            "RECIPE-DUMMY-VAL-10",
        ] == v

        v = pad_values(list("rocket"))
        assert v == [
            "r",
            "o",
            "c",
            "k",
            "e",
            "t",
            "RECIPE-DUMMY-VAL-1",
            "RECIPE-DUMMY-VAL-2",
            "RECIPE-DUMMY-VAL-3",
            "RECIPE-DUMMY-VAL-4",
            "RECIPE-DUMMY-VAL-5",
        ]

        v = pad_values(["a"], prefix="COW", bin_size=10)
        assert v == [
            "a",
            "COW1",
            "COW2",
            "COW3",
            "COW4",
            "COW5",
            "COW6",
            "COW7",
            "COW8",
            "COW9",
        ]


class AttrDictTestCase(RecipeTestCase):
    def test_attr_dict(self):
        d = AttrDict()
        assert isinstance(d, dict)
        d.foo = 2
        assert d["foo"] == 2
        d["bar"] = 3
        assert d.bar == 3


class FakerFormatterTestCase(RecipeTestCase):
    def test_formatter(self):
        formatter = FakerFormatter()
        # Faker providers can be accessed by attribute if they take no
        # arguments
        assert len(formatter.format("{fake:name}", fake=Faker())) > 0

        # They can no longer be accessed as callables
        with self.assertRaises(AttributeError):
            formatter.format("{fake:name()}", fake=Faker())

        # Some attributes can't be found
        with self.assertRaises(AttributeError):
            formatter.format("{fake:nm}", fake=Faker())

        # Parameterized values still work
        assert len(formatter.format("{fake:numerify|text=###}", fake=Faker())) == 3


class CowProvider(BaseProvider):
    def moo(self):
        return "moo"


class FakerAnonymizerTestCase(RecipeTestCase):
    def test_anonymizer_with_NO_params(self):
        a = FakerAnonymizer("{fake:random_uppercase_letter}")

        assert a("Value") == a("Value")
        assert a("boo") in uppercase

        b = FakerAnonymizer("{fake:military_apo}")
        assert b("Value") == b("Value")
        assert b("boo") == b("boo")
        assert b("Value") != b("boo")

    def test_anonymizer_with_params(self):
        a = FakerAnonymizer("{fake:numerify|text=###}")
        assert a("Value") == a("Value")

        b = FakerAnonymizer(
            "{fake:lexify|text=???,letters=abcdefghijklmnopqrstuvwxyz"
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ}"
        )
        assert len(b("value"))

        # Show we handle booleans
        before_today = FakerAnonymizer(
            "{fake:date_this_century|before_today=True,after_today=False}"
        )
        after_today = FakerAnonymizer(
            "{fake:date_this_century|before_today=False,after_today=True}"
        )

        # FakerAnonymizer always returns a string
        today = str(date.today())
        for let in "abcdefghijklmnopq":
            assert before_today(let) < today
            assert after_today(let) > today

    def test_anonymizer_with_postprocessor(self):
        # FakerAnonymizer always returns string unless converted
        a = FakerAnonymizer("{fake:ean8}")

        assert isinstance(a("Value"), str)

        b = FakerAnonymizer("{fake:ean8}", postprocessor=lambda x: int(x))

        assert isinstance(b("Value"), int)

        assert int(a("Value")) == b("Value")

    def test_anonymizer_with_provider(self):
        """Register a provider"""
        a = FakerAnonymizer("{fake:moo}", providers=[CowProvider])

        assert isinstance(a("Value"), str)
        assert a("Value") == "moo"

    def test_anonymizer_with_bad_providers(self):
        """Register a provider"""
        a = FakerAnonymizer("{fake:moo}", providers=[None, 4, CowProvider])

        assert isinstance(a("Value"), str)
        assert a("Value") == "moo"

    def test_anonymizer_with_stringprovider(self):
        """Register a string provider that is dynamically imported"""
        a = FakerAnonymizer("{fake:foo}", providers=["recipe.utils.TestProvider"])

        assert isinstance(a("Value"), str)
        assert a("Value") == "foo"

    def test_anonymizer_with_multipleproviders(self):
        """Register multiple providers"""
        a = FakerAnonymizer(
            "{fake:foo} {fake:moo}",
            providers=["recipe.utils.TestProvider", CowProvider],
        )

        assert isinstance(a("Value"), str)
        assert a("Value") == "foo moo"


class TestMakeSchema(RecipeTestCase):
    def test_make_schema(self):
        s = make_schema([])
        self.assertIsInstance(s, dict)
        self.assertEqual(
            s,
            {
                "schema": {
                    "metrics": {
                        "schema": {"type": "string", "required": True},
                        "type": "list",
                        "required": False,
                    },
                    "dimensions": {
                        "schema": {"type": "string", "required": True},
                        "type": "list",
                        "required": False,
                    },
                    "filters": {
                        "schema": {"type": "string", "required": True},
                        "type": "list",
                        "required": False,
                    },
                    "order_by": {
                        "schema": {"type": "string", "required": True},
                        "type": "list",
                        "required": False,
                    },
                },
                "type": "dict",
                "required": True,
            },
        )

    def test_make_schema_with_recipe_extensions(self):
        from recipe.extensions import AutomaticFilters

        # We can mixin schemas from extensions to get a combined schema
        s = make_schema([AutomaticFilters])
        self.assertIsInstance(s, dict)
        self.assertEqual(
            s,
            {
                "schema": {
                    "metrics": {
                        "schema": {"type": "string", "required": True},
                        "type": "list",
                        "required": False,
                    },
                    "dimensions": {
                        "schema": {"type": "string", "required": True},
                        "type": "list",
                        "required": False,
                    },
                    "filters": {
                        "schema": {"type": "string", "required": True},
                        "type": "list",
                        "required": False,
                    },
                    "order_by": {
                        "schema": {"type": "string", "required": True},
                        "type": "list",
                        "required": False,
                    },
                    "automatic_filters": {
                        "anyof": [
                            {"keyschema": {"type": "string"}, "type": "dict"},
                            {
                                "schema": {
                                    "keyschema": {"type": "string"},
                                    "type": "dict",
                                },
                                "type": "list",
                            },
                        ]
                    },
                    "include_automatic_filter_keys": {
                        "type": "list",
                        "schema": {"type": "string"},
                    },
                    "exclude_automatic_filter_keys": {
                        "type": "list",
                        "schema": {"type": "string"},
                    },
                    "apply_automatic_filters": {"type": "boolean"},
                    "strict_automatic_filters": {"type": "boolean"},
                },
                "type": "dict",
                "required": True,
            },
        )



from datetime import date, datetime

import structlog
from lark import GrammarError, Lark
from sqlalchemy import (
    func,
    text,
    cast,
    String,
    Date,
    DateTime,
    Integer,
    Float,
    Boolean,
    alias,
)
from typing import List, Optional

from .expression_grammar import (
    make_grammar,
    make_column_collection_for_selectable,
    make_column_collection_for_constant_literals,
    make_column_collection_for_constant_expressions,
    has_constant_expressions,
    has_constant_literals,
)
from .transformers import TransformToSQLAlchemyExpression
from .utils import mkkey
from .validators import SQLALchemyValidator

SLOG = structlog.get_logger(__name__)


LARK_CACHE = {}


class SQLAlchemyBuilder:
    @classmethod
    def get_builder(
        cls,
        selectable,
        *,
        constants: Optional[dict] = None,
        extra_selectables: Optional[List] = None,
        cache=None,
    ):
        return cls(
            selectable,
            constants=constants,
            extra_selectables=extra_selectables,
            cache=cache,
        )

    @classmethod
    def clear_builder_cache(cls):
        LARK_CACHE.clear()

    def __init__(
        self,
        selectable,
        *,
        constants: Optional[dict] = None,
        extra_selectables: Optional[List] = None,
        cache=None,
    ):
        """Parse a recipe field by building a custom grammar that
        uses the colums in a selectable.

        Args:
            selectable (Table): A SQLAlchemy selectable

        Keyword Args:
            extra_selectables (list): A list containing pairs of selectable,
              namespace string.
            cache (cache): An optional cache.
        """
        from recipe.core import Recipe

        if isinstance(selectable, Recipe):
            selectable = selectable.subquery()

        self.selectable = selectable
        # Database driver

        try:
            self.drivername = selectable.bind.url.drivername
        except AttributeError:
            try:
                self.drivername = selectable.metadata.bind.url.drivername
            except AttributeError:
                self.drivername = "unknown"

        self.cache = cache

        constants = constants or {}
        constant_expressions_cc = None

        # If we have expressions, we'll build a select statement
        # using the expressions, and make these into constants.
        # This limits constants to aggregate expressions on the base selectable.
        if has_constant_expressions(constants):
            self.columns = make_column_collection_for_selectable(selectable)
            self.finalize_grammar()

            constant_expressions_cc = make_column_collection_for_constant_expressions(
                self, constants, namespace="constants"
            )

        self.columns = make_column_collection_for_selectable(selectable)

        if constant_expressions_cc:
            self.columns.extend(constant_expressions_cc)

        # Add literal constants
        if has_constant_literals(constants):
            self.columns.extend(
                make_column_collection_for_constant_literals(
                    constants=constants, namespace="constants"
                )
            )

        if extra_selectables:
            for selectable, namespace in extra_selectables:
                self.columns.extend(
                    make_column_collection_for_selectable(
                        selectable, namespace=namespace
                    )
                )

        self.finalize_grammar()

    def finalize_grammar(self):
        """Once we have a set of columns, we can generate the parser and transformer"""
        self.grammar = make_grammar(self.columns)
        grammar_hash = mkkey("grammar", self.grammar)
        # Developer Note: cache key
        # This cache key is used for both LARK_CACHE as well as the SQLAlchemy
        # expressions that we use in `parse` below. This key must change any time the
        # table columns change, which it *should* do because the grammar contains
        # information about all types and columns in the table. If that ever changes
        # (e.g. we switch to separate steps for parsing and type/variable validation),
        # we'll need to calculate the expression hash key separately.
        self.cache_key = f"recipe-expr:{grammar_hash}"
        try:
            self.cached_trees = (
                self.cache.get(self.cache_key, {}) if self.cache is not None else None
            )
        except Exception:
            SLOG.exception("ingredient-cache-error")
            self.cached_trees = None
        if self.cache_key in LARK_CACHE:
            self.parser = LARK_CACHE[self.cache_key]
        else:
            # Constructing this Lark parser can take a significant amount of time (like,
            # nearly 1 second for some large tables), which is why we cache parsers in
            # LARK_CACHE. Unfortunately, by using an in-process cache, we are still
            # computing this redundantly quite often, because Juicebox processes get
            # cycled often and there are many workers in a given deployment.
            #
            # Lark supports serializing parsers to speed up loading, but unfortunately
            # it only supports this for LALR parsers, and we are using Earley.
            self.parser = Lark(
                self.grammar,
                parser="earley",
                ambiguity="resolve",
                start="col",
                propagate_positions=True,
                # predict_all=True,
            )
            LARK_CACHE[self.cache_key] = self.parser

        self.transformer = TransformToSQLAlchemyExpression(
            self.selectable, self.columns, self.drivername
        )

        # The data type of the last parsed expression
        self.last_datatype = None

    def parse(
        self,
        text,
        forbid_aggregation=False,
        enforce_aggregation=False,
        debug=False,
        convert_dates_with=None,
        convert_datetimes_with=None,
    ):
        """Return a parse tree for text

        Args:
            text (str): A field expression
            forbid_aggregation (bool, optional):
              The expression may not contain aggregations. Defaults to False.
            enforce_aggregation (bool, optional):
              Wrap the expression in an aggregation if one is not provided. Defaults to False.
            debug (bool, optional): Show some debug info. Defaults to False.
            convert_dates_with (str, optional): A converter to use for date fields
            convert_datetimes_with (str, optional): A converter to use for datetime fields

        Raises:
            GrammarError: A description of any errors and where they occur

        Returns:
            A tuple of
                ColumnElement: A SQLALchemy expression
                DataType: The datatype of the expression (bool, date, datetime, num, str)
        """
        key = mkkey(
            "parsed-ingredient",
            text,
            forbid_aggregation,
            enforce_aggregation,
            convert_dates_with,
            convert_datetimes_with,
        )
        cache_result = (
            self.cached_trees.get(key) if self.cached_trees is not None else None
        )

        extra_args = (
            key,
            enforce_aggregation,
            debug,
            convert_dates_with,
            convert_datetimes_with,
        )

        if cache_result is None:
            (tree, validator) = self._parse(text, forbid_aggregation)
            return self.tree_to_expression(tree, validator, *extra_args)
        else:
            (tree, validator) = cache_result
            try:
                return self.tree_to_expression(tree, validator, *extra_args)
            except Exception:
                SLOG.exception("cached-tree-to-validator-error")
                # If we get ANY error while dealing with the cached ingredient data, we
                # should just retry everything without using the cache. There could be
                # any number of things wrong with the cache, e.g. if it was produced on
                # an older version of Recipe (there are a lot of internal implementation
                # details encoded into the cached data).
                del self.cached_trees[key]
                (tree, validator) = self._parse(text, forbid_aggregation)
                return self.tree_to_expression(tree, validator, *extra_args)

    def _parse(self, text, forbid_aggregation):
        tree = self.parser.parse(text, start="col")
        validator = SQLALchemyValidator(text, forbid_aggregation, self.drivername)
        validator.visit(tree)
        return (tree, validator)

    def tree_to_expression(
        self,
        tree,
        validator,
        key,
        enforce_aggregation,
        debug,
        convert_dates_with,
        convert_datetimes_with,
    ):
        self.last_datatype = validator.last_datatype
        if validator.errors:
            if debug:
                print("".join(validator.errors))
                print("Tree:\n" + tree.pretty())
            raise GrammarError("".join(validator.errors))

        if debug:
            print("Tree:\n" + tree.pretty())
        self.transformer.text = text
        self.transformer.convert_dates_with = convert_dates_with
        self.transformer.convert_datetimes_with = convert_datetimes_with
        expr = self.transformer.transform(tree)

        # Wrap literal expressions in a cast so they can be labeled
        if isinstance(expr, (str, float, int, date, datetime, bool)):
            if isinstance(expr, str):
                self.last_datatype = "str"
                expr = cast(expr, String)
            elif isinstance(expr, date):
                self.last_datatype = "date"
                expr = cast(expr, Date)
            elif isinstance(expr, datetime):
                self.last_datatype = "datetime"
                expr = cast(expr, DateTime)
            elif isinstance(expr, int):
                self.last_datatype = "num"
                expr = cast(expr, Integer)
            elif isinstance(expr, float):
                self.last_datatype = "num"
                expr = cast(expr, Float)
            elif isinstance(expr, bool):
                self.last_datatype = "bool"
                expr = cast(expr, Boolean)
            else:
                raise GrammarError("Unsure of the datatype of {expr}")

        if (
            enforce_aggregation
            and not validator.found_aggregation
            and self.last_datatype == "num"
        ):
            result = (func.sum(expr), self.last_datatype)
        else:
            result = (expr, self.last_datatype)

        if self.cached_trees is not None:
            self.cached_trees[key] = (tree, validator)
        return result

    def save_cache(self):
        # see "Developer Note: cache key" for info about cache keys.
        try:
            self.cache.set(self.cache_key, self.cached_trees)
        except Exception:
            SLOG.exception("shelf-save-cache-error")



from functools import wraps


def recipe_arg(*args):
    """Decorator for recipe builder arguments.

    Promotes builder pattern by returning a recipe.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *_args, **_kwargs):
            from recipe import Recipe, RecipeExtension, BadRecipe

            if isinstance(self, Recipe):
                recipe = self
            elif isinstance(self, RecipeExtension):
                recipe = self.recipe
            else:
                raise BadRecipe(
                    "recipe_arg can only be applied to"
                    "methods of Recipe or RecipeExtension"
                )

            if recipe._query is not None:
                recipe.reset()

            func(self, *_args, **_kwargs)
            return recipe

        return wrapper

    return decorator



"""Convert parsed trees into SQLAlchemy objects """
from datetime import date
from .builders import SQLAlchemyBuilder
from lark.exceptions import GrammarError, LarkError

from recipe.exceptions import BadIngredient
from recipe.ingredients import InvalidIngredient

from .utils import ingredient_class_for_name


def _convert_bucket_to_field(field, bucket, buckets_default_label, builder):
    """Convert a bucket structure if statement

    This assumes that all conditions have been converted from
    partial conditions like '<5' to full conditions like 'age<5'
    with _convert_partial_conditions
    """

    def stringify(value):
        """Convert a value into a string that will parse"""
        if value is None:
            return "NULL"
        elif value is True:
            return "TRUE"
        elif value is False:
            return "FALSE"
        elif isinstance(value, str):
            return '"' + value.replace('"', '\\"') + '"'
        else:
            return str(value)

    parts, order_by_parts, idx = [], [], 0
    for itm in bucket:
        cond = itm.get("condition")
        label = itm.get("label")
        # Determine if the field needs to be prepended to the condition """"
        try:
            # The expression may not contain aggregations
            builder.parse(cond, forbid_aggregation=True)
        except Exception:
            cond = f"{field} {cond}"
            # Ensure this is parsable
            builder.parse(cond, forbid_aggregation=True)
        parts.append(cond)
        order_by_parts.append(cond)
        parts.append(stringify(label))
        order_by_parts.append(str(idx))
        idx += 1

    # Add the default value
    if buckets_default_label is None:
        buckets_default_label = "Not found"
    # Add a final default label
    parts.append(stringify(buckets_default_label))
    order_by_parts.append(str(9999))

    return "if(" + ",".join(parts) + ")", "if(" + ",".join(order_by_parts) + ")"


def set_date_aggregation_from_format(ingr_dict: dict):
    """
    Determine the date aggregation from a format property.

    Allowed values are:

    year: Group dates and datetimes by the first day of the year.
    month: Group dates and datestimes by the first day of the month.
    day: Group by the day (this is only relevant to datetimes)

    This will wrap any expression with an appropriate date trunctation
    """
    # Note: If date_aggregation is populated on ingr_dict
    # we can remove this hackito.
    if "date_aggregation" in ingr_dict or "format" not in ingr_dict:
        return

    # Derive date_aggregation from the format property. This is not a property
    # that recipe cares about.
    clean_format = ingr_dict["format"]
    # The format string may be wrapped with d3 time formatting brackets <>
    if clean_format.startswith("<") and clean_format.endswith(">"):
        clean_format = clean_format[1:-1]

    if clean_format == "%Y":
        ingr_dict["date_aggregation"] = "year"
    elif (
        "%B" in clean_format
        and "%Y" in clean_format
        and "%-d" not in clean_format
        and "%d" not in clean_format
    ):
        ingr_dict["date_aggregation"] = "month"
    elif "%H" not in clean_format:
        ingr_dict["date_aggregation"] = "day"


def convert_quickselects(
    builder: SQLAlchemyBuilder, ingr_dict: dict, builder_kwargs: dict
):
    """Convert quickselects from expressions to an object with name, expression"""
    if "quickselects" in ingr_dict:
        parsed_quickselects = []
        for qs in ingr_dict.pop("quickselects", []):
            condition_defn = qs.get("condition")
            expr, _ = builder.parse(
                condition_defn, forbid_aggregation=True, **builder_kwargs
            )
            parsed_quickselects.append({"name": qs["name"], "condition": expr})
        ingr_dict["quickselects"] = parsed_quickselects


def convert_filter(builder: SQLAlchemyBuilder, ingr_dict: dict, builder_kwargs: dict):
    """If a filter property exists, validate that it is a boolean expression and add it
    to the ingredient filters. Silently ignore if we return the wrong datatype."""
    if "filter" in ingr_dict:
        filt_expression = ingr_dict.pop("filter")

        expr, datatype = builder.parse(
            filt_expression, forbid_aggregation=True, **builder_kwargs
        )
        if datatype == "bool":
            ingr_dict["filters"] = [expr]
        else:
            raise BadIngredient(
                f"filter: '{filt_expression}' has data type '{datatype}'. It must must be boolean."
            )


def convert_buckets_to_field_defn(
    builder: SQLAlchemyBuilder, ingr_dict: dict, fld_defn: str, builder_kwargs: dict
):
    """If a buckets key exists, convert it to an expression, add
    the order_by to extra_fields.
    """
    buckets = ingr_dict.pop("buckets", None)
    buckets_default_label = ingr_dict.pop("buckets_default_label", None)
    if buckets:
        fld_defn, order_by_fld = _convert_bucket_to_field(
            fld_defn, buckets, buckets_default_label, builder
        )
        if "extra_fields" not in ingr_dict:
            ingr_dict["extra_fields"] = []
        ingr_dict["extra_fields"].append(
            {"name": "order_by_expression", "field": order_by_fld}
        )
    return fld_defn


def convert_extra_fields(
    builder: SQLAlchemyBuilder, ingr_dict: dict, builder_kwargs: dict
):
    """
    Dimensions may contain extra expressions in the extra_fields list.

    Convert extra fields to sqlalchemy expressions and add them directly to
    the kwargs, saving datatypes in datatype_by_role
    """
    for extra_fld in ingr_dict.pop("extra_fields", []):
        # Extra_fields will be dicts with two keys
        # name and field
        raw_role = extra_fld.get("name")
        role = raw_role.rstrip("_expression")

        expr, datatype = builder.parse(
            extra_fld.get("field"), forbid_aggregation=True, **builder_kwargs
        )
        ingr_dict["datatype_by_role"]["role"] = datatype
        ingr_dict[raw_role] = expr


def create_ingredient_from_parsed(
    ingr_config: dict, builder: SQLAlchemyBuilder, debug: bool = False
):
    """Create an ingredient from config version 2 object ."""
    kind = ingr_config.pop("kind", "metric")
    IngredientClass = ingredient_class_for_name(kind.title())
    if IngredientClass is None:
        raise BadIngredient(f"Unknown ingredient kind {kind}")

    args = []

    clean_format = ingr_config.get("format")
    if (
        isinstance(clean_format, str)
        and clean_format.startswith("<")
        and clean_format.endswith(">")
    ):
        clean_format = clean_format[1:-1]

    # TODO: this can be removed when "date_aggregation" is set on ingr_dict directly.
    set_date_aggregation_from_format(ingr_dict=ingr_config)

    # For some formats, we will automatically convert dates to year or month in the builder
    date_aggregation = ingr_config.get("date_aggregation")
    date_aggr_lookup = {"year": "year_conv", "month": "month_conv"}
    dt_aggr_lookup = {
        "year": "dt_year_conv",
        "month": "dt_month_conv",
        "day": "dt_day_conv",
    }

    builder_kwargs = {
        "debug": debug,
        "convert_dates_with": date_aggr_lookup.get(date_aggregation),
        "convert_datetimes_with": dt_aggr_lookup.get(date_aggregation),
    }

    if builder.drivername.startswith("mssql") or builder.drivername.startswith(
        "snowflake"
    ):
        # SQLServer can not use aliases in group bys and also
        # does not support date/time conversions due to an issue with pyodbc
        # parameters in queries
        # https://github.com/mkleehammer/pyodbc/issues/479
        default_group_by_strategy = "direct"
    else:
        default_group_by_strategy = "labels"

    try:
        if kind == "metric":
            fld_defn = ingr_config.pop("field", None)
            # SQLAlchemy ingredient with required aggregation
            expr, datatype = builder.parse(
                fld_defn, enforce_aggregation=True, **builder_kwargs
            )
            # Save the data type in the ingredient
            ingr_config["datatype"] = datatype

            convert_filter(builder, ingr_config, builder_kwargs)
            convert_quickselects(builder, ingr_config, builder_kwargs)
            args = [expr]

        elif kind == "dimension":
            ingr_config["group_by_strategy"] = ingr_config.get(
                "group_by_strategy", default_group_by_strategy
            )

            fld_defn = ingr_config.pop("field", None)
            fld_defn = convert_buckets_to_field_defn(
                builder, ingr_config, fld_defn, builder_kwargs
            )

            expr, datatype = builder.parse(
                fld_defn, forbid_aggregation=True, **builder_kwargs
            )
            # Save the data type in the ingredient
            ingr_config["datatype"] = datatype
            args = [expr]
            ingr_config["datatype_by_role"] = {"value": datatype}

            convert_extra_fields(builder, ingr_config, builder_kwargs)
            convert_filter(builder, ingr_config, builder_kwargs)
            convert_quickselects(builder, ingr_config, builder_kwargs)

        elif kind == "filter":
            condition_defn = ingr_config.get("condition")
            expr, datatype = builder.parse(
                condition_defn, forbid_aggregation=True, **builder_kwargs
            )
            args = [expr]

        elif kind == "having":
            condition_defn = ingr_config.get("condition")
            expr, datatype = builder.parse(
                condition_defn, forbid_aggregation=False, **builder_kwargs
            )
            args = [expr]

    except (GrammarError, LarkError) as e:
        error_msg = str(e)
        if "Expecting:" in error_msg:
            error_msg = error_msg.split("Expecting:")[0]

        error = {"type": "Can not parse field", "extra": {"details": error_msg}}
        return InvalidIngredient(error=error)

    try:
        return IngredientClass(*args, **ingr_config)
    except BadIngredient as e:
        # Some internal error while running the Ingredient constructor
        error = {"type": "bad_ingredient", "extra": {"details": str(e)}}
        return InvalidIngredient(error=error)



from sqlalchemy.sql import expression, func, distinct, text
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.types import Numeric, String, Integer

# A custom age function for postgres


class postgres_age(expression.FunctionElement):
    type = Numeric()
    name = "postgres_age"


@compiles(postgres_age)
def pgage(element, compiler, **kw):
    # Calculate the difference in years, then adjust based on whether the current date has passed
    # using EXTRACT('dayofyear') failed on edge cases
    clauses = compiler.process(element.clauses)
    return (
        "DATEDIFF('YEAR', %s, CURRENT_DATE) - " % clauses
        + "CASE WHEN extract('month' from CURRENT_DATE) + extract('day' from CURRENT_DATE)/100.0 "
        + "< extract('month' from %s)+ extract('day' from %s)/100.0 THEN 1 ELSE 0 END"
        % (clauses, clauses)
    )


# Custom function definitions for bigquery


class bq_median(expression.FunctionElement):
    type = Numeric()
    name = "bq_median"


class bq_percentile1(expression.FunctionElement):
    type = Numeric()
    name = "bq_percentile1"


class bq_percentile5(expression.FunctionElement):
    type = Numeric()
    name = "bq_percentile5"


class bq_percentile10(expression.FunctionElement):
    type = Numeric()
    name = "bq_percentile10"


class bq_percentile25(expression.FunctionElement):
    type = Numeric()
    name = "bq_percentile25"


class bq_percentile75(expression.FunctionElement):
    type = Numeric()
    name = "bq_percentile75"


class bq_percentile90(expression.FunctionElement):
    type = Numeric()
    name = "bq_percentile90"


class bq_percentile95(expression.FunctionElement):
    type = Numeric()
    name = "bq_percentile95"


class bq_percentile99(expression.FunctionElement):
    type = Numeric()
    name = "bq_percentile99"


@compiles(bq_median, "bigquery")
def bqmedian(element, compiler, **kw):
    return f"approx_quantiles({compiler.process(element.clauses)}, 2)[OFFSET(1)]"


@compiles(bq_percentile1, "bigquery")
def bqpercentile1(element, compiler, **kw):
    return f"approx_quantiles({compiler.process(element.clauses)}, 100)[OFFSET(1)]"


@compiles(bq_percentile5, "bigquery")
def bqpercentile5(element, compiler, **kw):
    return f"approx_quantiles({compiler.process(element.clauses)}, 20)[OFFSET(1)]"


@compiles(bq_percentile10, "bigquery")
def bqpercentile10(element, compiler, **kw):
    return f"approx_quantiles({compiler.process(element.clauses)}, 10)[OFFSET(1)]"


@compiles(bq_percentile25, "bigquery")
def bqpercentile25(element, compiler, **kw):
    return f"approx_quantiles({compiler.process(element.clauses)}, 4)[OFFSET(1)]"


@compiles(bq_percentile75, "bigquery")
def bqpercentile75(element, compiler, **kw):
    return f"approx_quantiles({compiler.process(element.clauses)}, 4)[OFFSET(3)]"


@compiles(bq_percentile90, "bigquery")
def bqpercentile90(element, compiler, **kw):
    return f"approx_quantiles({compiler.process(element.clauses)}, 10)[OFFSET(9)]"


@compiles(bq_percentile95, "bigquery")
def bqpercentile95(element, compiler, **kw):
    return f"approx_quantiles({compiler.process(element.clauses)}, 20)[OFFSET(19)]"


@compiles(bq_percentile99, "bigquery")
def bqpercentile99(element, compiler, **kw):
    return f"approx_quantiles({compiler.process(element.clauses)}, 100)[OFFSET(99)]"


# An age calculation for bigquery


class bq_age(expression.FunctionElement):
    type = Numeric()
    name = "bq_age"


@compiles(bq_age, "bigquery")
def bqage(element, compiler, **kw):
    clauses = compiler.process(element.clauses)
    return (
        f"DATE_DIFF(CURRENT_DATE, {clauses}, YEAR) - IF(EXTRACT(MONTH FROM CURRENT_DATE) + EXTRACT(DAY FROM CURRENT_DATE)/100.0 "
        + f"< EXTRACT(MONTH FROM {clauses})+ EXTRACT(DAY FROM {clauses})/100.0, 1, 0)"
    )


####################
# Aggregations are a callable on a column expressoin that yields an
# aggregated column expression
# for instance sum(sales) => func.sum(MyTable.sales)
# Different database engines can use different aggregations
####################

aggregations = {
    "sum": func.sum,
    "min": func.min,
    "max": func.max,
    "avg": func.avg,
    "count": func.count,
    "count_distinct": lambda fld: func.count(distinct(fld)),
    # Technically "none" is not an aggregation but we're keeping
    # it here for backward compatibility
    "none": lambda fld: fld,
    None: lambda fld: fld,
    # Percentile aggregations do not work in all engines
    "median": func.median,
    "percentile1": lambda fld: func.percentile_cont(0.01).within_group(fld),
    "percentile5": lambda fld: func.percentile_cont(0.05).within_group(fld),
    "percentile10": lambda fld: func.percentile_cont(0.10).within_group(fld),
    "percentile25": lambda fld: func.percentile_cont(0.25).within_group(fld),
    "percentile50": lambda fld: func.percentile_cont(0.50).within_group(fld),
    "percentile75": lambda fld: func.percentile_cont(0.75).within_group(fld),
    "percentile90": lambda fld: func.percentile_cont(0.90).within_group(fld),
    "percentile95": lambda fld: func.percentile_cont(0.95).within_group(fld),
    "percentile99": lambda fld: func.percentile_cont(0.99).within_group(fld),
}

# Additional aggregations supported on redshift
aggregations_redshift = {
    # Percentile aggregations do not work in all engines
    "median": func.median,
    "percentile1": lambda fld: func.percentile_cont(0.01).within_group(fld),
    "percentile5": lambda fld: func.percentile_cont(0.05).within_group(fld),
    "percentile10": lambda fld: func.percentile_cont(0.10).within_group(fld),
    "percentile25": lambda fld: func.percentile_cont(0.25).within_group(fld),
    "percentile50": lambda fld: func.percentile_cont(0.50).within_group(fld),
    "percentile75": lambda fld: func.percentile_cont(0.75).within_group(fld),
    "percentile90": lambda fld: func.percentile_cont(0.90).within_group(fld),
    "percentile95": lambda fld: func.percentile_cont(0.95).within_group(fld),
    "percentile99": lambda fld: func.percentile_cont(0.99).within_group(fld),
}


aggregations_bigquery = {
    "median": lambda fld: bq_median(fld),
    "percentile1": lambda fld: bq_percentile1(fld),
    "percentile5": lambda fld: bq_percentile5(fld),
    "percentile10": lambda fld: bq_percentile10(fld),
    "percentile25": lambda fld: bq_percentile25(fld),
    "percentile50": lambda fld: bq_median(fld),
    "percentile75": lambda fld: bq_percentile75(fld),
    "percentile90": lambda fld: bq_percentile90(fld),
    "percentile95": lambda fld: bq_percentile95(fld),
    "percentile99": lambda fld: bq_percentile99(fld),
}

# A dictionary of aggregations keyed by sqlalchemy drivername
# 'default' are
aggregations_by_engine = {
    "default": aggregations,
    "redshift+psycopg2": aggregations_redshift,
    "bigquery": aggregations_bigquery,
}


#######################
# Conversions are a callable on a column expression that yields a
# nonaggregated column expression
# for instance, quarter(sales_date) => func.date_trunc('quarter', MyTable.sales_date)
#######################

conversions = {
    "month": lambda fld: func.date_trunc("month", fld),
    "week": lambda fld: func.date_trunc("week", fld),
    "year": lambda fld: func.date_trunc("year", fld),
    "quarter": lambda fld: func.date_trunc("quarter", fld),
    "string": lambda fld: func.cast(fld, String()),
    "int": lambda fld: func.cast(fld, Integer()),
}


conversions_redshift = {
    # age doesn't work on all databases
    "age": lambda fld: postgres_age(fld)
}

conversions_bigquery = {
    "month": lambda fld: func.date_trunc(fld, text("month")),
    "week": lambda fld: func.date_trunc(fld, text("week")),
    "year": lambda fld: func.date_trunc(fld, text("year")),
    "quarter": lambda fld: func.date_trunc(fld, text("quarter")),
    "age": lambda fld: bq_age(fld),
}


# A dictionary of conversions keyed by sqlalchemy drivername
conversions_by_engine = {
    "default": conversions,
    "redshift+psycopg2": conversions_redshift,
    "bigquery": conversions_bigquery,
}



from copy import copy

import pytest
from sqlalchemy import join
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.elements import ColumnElement
from yaml import safe_load

from recipe import (
    AutomaticShelf,
    InvalidColumnError,
    BadRecipe,
    Dimension,
    Metric,
    Filter,
    Shelf,
)
import recipe
from recipe.ingredients import Ingredient, InvalidIngredient
from recipe.shelf import introspect_table
from recipe.schemas.utils import find_column

from .test_base import RecipeTestCase


class FindColumnTestCase(RecipeTestCase):
    def test_find_column_from_recipe(self):
        """Can find columns in a recipe."""
        content = """
        state:
            kind: Dimension
            field: state
        sex:
            kind: Dimension
            field: sex
        age:
            kind: Dimension
            field: age
        sum_pop2000:
            kind: Metric
            field: pop2000
        pop2008:
            kind: Metric
            field: pop2008
        ttlpop:
            kind: Metric
            field: pop2000 + pop2008

        """
        shelf = Shelf.from_yaml(content, self.census_table)
        recipe = self.recipe(shelf=shelf).metrics("sum_pop2000").dimensions("state")
        assert (
            recipe.to_sql()
            == """SELECT census.state AS state,
       sum(census.pop2000) AS sum_pop2000
FROM census
GROUP BY state"""
        )

        col = find_column(recipe, "state")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))

        with self.assertRaises(InvalidColumnError):
            find_column(recipe, "census_state")
        with self.assertRaises(InvalidColumnError):
            find_column(recipe, "foo")

        col = find_column(recipe, "sum_pop2000")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))

        with self.assertRaises(InvalidColumnError):
            find_column(recipe, "pop2000")

    def test_find_column_from_table_mytable(self):
        """SQLALchemy ORM Tables can be used and return
        InstrumentedAttributes"""
        col = find_column(self.basic_table, "first")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))

        col2 = find_column(self.basic_table, "foo_first")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))
        self.assertEqual(col, col2)

        col = find_column(self.basic_table, "last")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))

        col = find_column(self.basic_table, "age")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))

        col = find_column(self.basic_table, "foo_age")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))

        with self.assertRaises(InvalidColumnError):
            find_column(self.basic_table, "foo")

    def test_find_column_from_join(self):
        """Columns can be found in a join"""
        j = join(
            self.census_table,
            self.state_fact_table,
            self.census_table.c.state == self.state_fact_table.c.name,
        )

        # Names can be either the column name or the {tablename}_{column}
        col = find_column(j, "state")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))

        col2 = find_column(j, "census_state")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))
        self.assertEqual(col, col2)

        # Names can be either the column name or the {tablename}_{column}
        col = find_column(j, "sex")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))

        col2 = find_column(j, "census_sex")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))
        self.assertEqual(col, col2)

        col = find_column(j, "assoc_press")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))

        # Columns can be referenced as {tablename}_{column}
        col2 = find_column(j, "state_fact_assoc_press")
        assert isinstance(col, (ColumnElement, InstrumentedAttribute))
        self.assertEqual(col, col2)

        with self.assertRaises(InvalidColumnError):
            find_column(j, "foo")

    def test_find_column_from_invalid_type(self):
        """Columns can be found in a join"""
        with self.assertRaises(InvalidColumnError):
            find_column(1, "foo")
        with self.assertRaises(InvalidColumnError):
            find_column(self.basic_table.c.first, "foo")


class ShelfConstructionTestCase(RecipeTestCase):
    def test_pass_some_metadata(self):
        shelf = Shelf(metadata={"a": "hello"})
        self.assertEqual(shelf.Meta.metadata["a"], "hello")

    def test_Meta_is_not_shared(self):
        shelf = Shelf(metadata={"a": "hello"})
        shelf2 = Shelf(metadata={"b": "there"})
        self.assertEqual(shelf.Meta.metadata, {"a": "hello"})
        self.assertEqual(shelf2.Meta.metadata, {"b": "there"})


class ShelfTestCase(RecipeTestCase):
    def setUp(self):
        super().setUp()
        # Be sure we don't modify the shelf while testing
        self.shelf = copy(self.shelf)

    def test_find(self):
        """Find ingredients on the shelf"""
        ingredient = self.shelf.find("first", Dimension)
        self.assertEqual(ingredient.id, "first")

        # Raise if the wrong type
        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("first", Metric)

        # Raise if key not present in shelf
        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("foo", Dimension)

        # Raise if key is not an ingredient or string
        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find(2.0, Dimension)

        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("foo", Dimension)

        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find(2.0, Dimension)

        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("foo", Dimension)

        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("foo", Dimension)

        self.shelf["foo"] = Dimension(self.basic_table.c.last)
        ingredient = self.shelf.find("last", Dimension)
        self.assertEqual(ingredient.id, "last")

    def test_find_filter(self):
        self.shelf["age_gt_20"] = Filter(self.basic_table.c.age > 20)

        ingredient = self.shelf.find("age_gt_20", Filter)
        self.assertEqual(ingredient.id, "age_gt_20")

        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("age_gt_20", Dimension)

    def test_repr(self):
        """Find ingredients on the shelf"""
        print(self.shelf.__repr__())
        assert (
            self.shelf.__repr__()
            == """(Dimension)first foo.first
(Dimension)firstlast foo.first foo.last
(Dimension)last foo.last
(Metric)age sum(foo.age)"""
        )

    def test_keys(self):
        self.assertEqual(
            sorted(self.shelf.keys()), ["age", "first", "firstlast", "last"]
        )

    def test_update(self):
        """Shelves can be updated with other shelves"""
        new_shelf = Shelf({"squee": Dimension(self.basic_table.c.first)})
        self.assertEqual(len(self.shelf), 4)
        self.shelf.update(new_shelf)
        self.assertEqual(len(self.shelf), 5)

    def test_update_key_value(self):
        """Shelves can be built with key_values and updated"""
        new_shelf = Shelf(squee=Dimension(self.basic_table.c.first))
        self.assertEqual(len(self.shelf), 4)
        self.shelf.update(new_shelf)
        self.assertEqual(len(self.shelf), 5)
        assert isinstance(self.shelf.get("squee"), Dimension)

    def test_update_key_value_direct(self):
        """Shelves can be updated directly with key_value"""
        self.assertEqual(len(self.shelf), 4)
        self.shelf.update(squee=Dimension(self.basic_table.c.first))
        self.assertEqual(len(self.shelf), 5)
        assert isinstance(self.shelf.get("squee"), Dimension)

    def test_brew(self):
        recipe_parts = self.shelf.brew_query_parts()
        self.assertEqual(len(recipe_parts["columns"]), 5)
        self.assertEqual(len(recipe_parts["group_bys"]), 4)
        self.assertEqual(len(recipe_parts["filters"]), 0)
        self.assertEqual(len(recipe_parts["havings"]), 0)

    def test_anonymize(self):
        """We can save and store anonymization context"""
        assert self.shelf.Meta.anonymize is False
        self.shelf.Meta.anonymize = True
        assert self.shelf.Meta.anonymize is True

    def test_get(self):
        """Find ingredients on the shelf"""
        ingredient = self.shelf["first"]
        self.assertEqual(ingredient.id, "first")

        ingredient = self.shelf.get("first", None)
        self.assertEqual(ingredient.id, "first")

        ingredient = self.shelf.get("primo", None)
        assert ingredient is None

    def test_get_doesnt_mutate(self):
        """
        Sharing ingredients between shelves won't cause race conditions on
        their `.id` and `.anonymize` attributes.
        """
        ingr = Ingredient(id="b")
        shelf = Shelf({"a": ingr})
        self.assertEqual(shelf["a"].id, "a")
        self.assertEqual(ingr.id, "b")

    def test_add_to_shelf(self):
        """We can add an ingredient to a shelf"""
        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("foo", Dimension)

        self.shelf["foo"] = Dimension(self.basic_table.c.last)
        ingredient = self.shelf.find("last", Dimension)
        self.assertEqual(ingredient.id, "last")

    def test_setitem_type_error(self):
        """Only ingredients can be added to shelves."""
        with self.assertRaises(TypeError):
            self.shelf["foo"] = 3

    def test_use_type_error(self):
        """`use` requires Ingredients"""
        with self.assertRaises(TypeError):
            self.shelf.use(3)

    def test_clear(self):
        self.shelf = copy(self.shelf)
        self.assertEqual(len(self.shelf), 4)
        self.shelf.clear()
        self.assertEqual(len(self.shelf), 0)

    def test_dimension_ids(self):
        self.assertEqual(len(self.shelf), 4)
        self.assertEqual(len(self.shelf.dimension_ids), 3)
        self.assertEqual(
            sorted(self.shelf.dimension_ids), ["first", "firstlast", "last"]
        )

    def test_metric_ids(self):
        self.assertEqual(len(self.shelf.metric_ids), 1)
        self.assertEqual(self.shelf.metric_ids, ("age",))

    def test_filter_ids(self):
        self.assertEqual(len(self.shelf.filter_ids), 0)


class ShelfFromYamlTestCase(RecipeTestCase):
    def make_shelf(self, content, table=None):
        if table is None:
            table = self.basic_table
        self.shelf = Shelf.from_yaml(content, table)
        self.shelf.Meta.anonymize = False

    def setUp(self):
        super().setUp()
        self.make_shelf(
            """
first:
    kind: Dimension
    field: first
last:
    kind: Dimension
    field: last
age:
    kind: Metric
    field: age
"""
        )

    def test_find(self):
        """Find ingredients on the shelf"""
        ingredient = self.shelf.find("first", Dimension)
        self.assertEqual(ingredient.id, "first")

        # Raise if the wrong type
        with self.assertRaises(Exception):
            ingredient = self.shelf.find("first", Metric)

        # Raise if key not present in shelf
        with self.assertRaises(Exception):
            ingredient = self.shelf.find("foo", Dimension)

        # Raise if key is not an ingredient or string
        with self.assertRaises(Exception):
            ingredient = self.shelf.find(2.0, Dimension)

        with self.assertRaises(Exception):
            ingredient = self.shelf.find("foo", Dimension)

        with self.assertRaises(Exception):
            ingredient = self.shelf.find(2.0, Dimension)

        with self.assertRaises(Exception):
            ingredient = self.shelf.find("foo", Dimension)

        with self.assertRaises(Exception):
            ingredient = self.shelf.find("foo", Dimension)

        self.shelf["foo"] = Dimension(self.basic_table.c.last)
        ingredient = self.shelf.find("last", Dimension)
        self.assertEqual(ingredient.id, "last")

    def test_repr(self):
        """Find ingredients on the shelf"""
        assert (
            self.shelf.__repr__()
            == """(Dimension)first foo.first
(Dimension)last foo.last
(Metric)age sum(foo.age)"""
        )

    def test_update(self):
        """Shelves can be updated with other shelves"""
        new_shelf = Shelf({"squee": Dimension(self.basic_table.c.first)})
        self.assertEqual(len(self.shelf), 3)
        self.shelf.update(new_shelf)
        self.assertEqual(len(self.shelf), 4)

    def test_update_key_value(self):
        """Shelves can be built with key_values and updated"""
        new_shelf = Shelf(squee=Dimension(self.basic_table.c.first))
        self.assertEqual(len(self.shelf), 3)
        self.shelf.update(new_shelf)
        self.assertEqual(len(self.shelf), 4)
        assert isinstance(self.shelf.get("squee"), Dimension)

    def test_update_key_value_direct(self):
        """Shelves can be updated directly with key_value"""
        self.assertEqual(len(self.shelf), 3)
        self.shelf.update(squee=Dimension(self.basic_table.c.first))
        self.assertEqual(len(self.shelf), 4)
        assert isinstance(self.shelf.get("squee"), Dimension)

    def test_brew(self):
        recipe_parts = self.shelf.brew_query_parts()
        self.assertEqual(len(recipe_parts["columns"]), 3)
        self.assertEqual(len(recipe_parts["group_bys"]), 2)
        self.assertEqual(len(recipe_parts["filters"]), 0)
        self.assertEqual(len(recipe_parts["havings"]), 0)

    def test_anonymize(self):
        """We can save and store anonymization context"""
        assert self.shelf.Meta.anonymize is False
        self.shelf.Meta.anonymize = True
        assert self.shelf.Meta.anonymize is True

    def test_anonymize_keeps_ingredients_up_to_date(self):
        """Setting the anonymize attribute causes all ingredients to be
        updated.
        """
        assert self.shelf["first"].anonymize is False
        self.shelf.Meta.anonymize = True
        assert self.shelf["first"].anonymize is True

    def test_get(self):
        """Find ingredients on the shelf"""
        ingredient = self.shelf["first"]
        self.assertEqual(ingredient.id, "first")

        ingredient = self.shelf.get("first", None)
        self.assertEqual(ingredient.id, "first")

        ingredient = self.shelf.get("primo", None)
        assert ingredient is None

    def test_add_to_shelf(self):
        """We can add an ingredient to a shelf"""
        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("foo", Dimension)

        self.shelf["foo"] = Dimension(self.basic_table.c.last)
        ingredient = self.shelf.find("last", Dimension)
        self.assertEqual(ingredient.id, "last")

    def test_clear(self):
        self.assertEqual(len(self.shelf), 3)
        self.shelf.clear()
        self.assertEqual(len(self.shelf), 0)

    def test_invalid_kind(self):
        content = """
oldage:
    kind: Metric2
    field:
        value: age
"""
        with self.assertRaises(Exception):
            self.make_shelf(content)


class ShelfFromValidatedYamlTestCase(ShelfFromYamlTestCase):
    """Test that shelves are created correctly using
    sureberus validation.
    """

    def make_shelf(self, content, table=None):
        if table is None:
            table = self.basic_table
        self.shelf = Shelf.from_validated_yaml(content, table)
        self.shelf.Meta.anonymize = False

    def test_invalid_definition(self):
        content = """
oldage:
    kind: Metric
    field: age
invalid:
    kind: Dimension
    field: age
    # raw_field is a reserved name and can't be assigned in config.
    raw_field: first
"""
        self.make_shelf(content)
        assert isinstance(self.shelf["oldage"], Metric)
        assert isinstance(self.shelf["invalid"], InvalidIngredient)
        assert (
            self.shelf["invalid"].error["extra"]["details"]
            == "raw is a reserved role in dimensions"
        )


class ShelfFromConfigTestCase(ShelfFromValidatedYamlTestCase):
    def make_shelf(self, content, table=None):
        if table is None:
            table = self.basic_table
        obj = safe_load(content)
        self.shelf = Shelf.from_config(obj, table)


class TestShelfFromIntrospection(RecipeTestCase):
    """Test that shelves are created correctly using
    sureberus validation.
    """

    def test_shelf(self):
        """Cerberus validated shelf doesn't accept null."""
        content = """
state:
    kind: Dimension
    field: state
sex:
    kind: Dimension
    field: sex
age:
    kind: Dimension
    field: age
pop2000:
    kind: Metric
    field: pop2000
pop2008:
    kind: Metric
    field: pop2008
ttlpop:
    kind: Metric
    field: pop2000 + pop2008

"""
        shelf = Shelf.from_validated_yaml(content, self.census_table)

        recipe = self.recipe(shelf=shelf).metrics("pop2000").dimensions("state")
        assert (
            recipe.to_sql()
            == """SELECT census.state AS state,
       sum(census.pop2000) AS pop2000
FROM census
GROUP BY state"""
        )
        assert (
            recipe.dataset.tsv
            == """state	pop2000	state_id\r
Tennessee	5685230	Tennessee\r
Vermont	609480	Vermont\r
"""
        )


class AutomaticShelfTestCase(RecipeTestCase):
    def setUp(self):
        super().setUp()
        self.shelf = AutomaticShelf(self.basic_table)

    def test_auto_find(self):
        """Find ingredients on the shelf"""
        ingredient = self.shelf.find("first", Dimension)
        self.assertEqual(ingredient.id, "first")

        # Raise if the wrong type
        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("first", Metric)

        # Raise if key not present in shelf
        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find("foo", Dimension)

        # Raise if key is not an ingredient or string
        with self.assertRaises(BadRecipe):
            ingredient = self.shelf.find(2.0, Dimension)

    def test_get(self):
        """Find ingredients on the shelf"""
        ingredient = self.shelf["first"]
        self.assertEqual(ingredient.id, "first")

        ingredient = self.shelf.get("first", None)
        self.assertEqual(ingredient.id, "first")

        ingredient = self.shelf["age"]
        self.assertEqual(str(ingredient.columns[0]), "sum(foo.age)")

        ingredient = self.shelf.get("primo", None)
        assert ingredient is None

    def test_introspect_table(self):
        config = introspect_table(self.basic_table)
        self.assertEqual(
            config,
            {
                "age": {"field": "age", "kind": "Metric"},
                "first": {"field": "first", "kind": "Dimension"},
                "last": {"field": "last", "kind": "Dimension"},
            },
        )



import abc
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from recipe import SETTINGS


class OvenBase(metaclass=abc.ABCMeta):
    """Base class for ovens"""

    def __init__(self, connection_string=None):
        self.engine = self.init_engine(connection_string)
        self.Session = self.init_session()

    @abc.abstractmethod
    def init_engine(self, connection_string=None, **kwargs):
        """Initializes a SQLAlchemy Engine for a given connection string with
        all other keyword arguments passed to the create_engine function. The
        connection uses pre-ping to verify connections.

        :param self: a reference to ourselves
        :param connection_string: a reference to ourselves
        :param kwargs: a collection of arguments passed to the engine
        :type self: Oven
        :type connection_string: str
        :type kwargs: dict
        :return: A SQLAlchemy Engine with connection checking
        :rtype: SQLAlchemy.Engine
        """
        if not connection_string:
            return

        connection_settings = {
            "pool_size": SETTINGS.POOL_SIZE,
            "pool_recycle": SETTINGS.POOL_RECYCLE,
            "pool_pre_ping": True,
        }
        connection_settings.update(kwargs)

        engine = create_engine(connection_string, **connection_settings)
        return engine

    @abc.abstractmethod
    def init_session(self):
        """Initializes a SQLAlchemy Session with the Oven's engine

        :param self: a reference to ourselves
        :type self: Oven
        :return: A SQLAlchemy Session using self.engine
        :rtype: SQLAlchemy.Session
        """
        if not self.engine:
            return

        return sessionmaker(bind=self.engine)



import csv
import io
import os
from datetime import date
from typing import Iterator
from unittest import TestCase

from dateutil.relativedelta import relativedelta
from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    Float,
    Integer,
    MetaData,
    String,
    Table,
    distinct,
    func,
    select,
)
from yaml import safe_load

from recipe import Dimension, Filter, IdValueDimension, Metric, Recipe, Shelf, get_oven


def strip_columns_from_csv(content: str, ignore_columns: list = None) -> str:
    if ignore_columns:
        content = str_dedent(content)
        rows = list(csv.DictReader(content.splitlines()))
        for row in rows:
            for col in ignore_columns:
                row.pop(col, None)
        if rows:
            csv_content = io.StringIO()
            first_row = rows[0]
            writer = csv.DictWriter(csv_content, fieldnames=list(first_row.keys()))
            writer.writeheader()
            writer.writerows(rows)
            return csv_content.getvalue().replace("\r\n", "\n").strip("\n")

    return content


def str_dedent(s: str) -> str:
    return "\n".join([x.lstrip() for x in s.split("\n")]).lstrip("\n").rstrip("\n")


class UtilsTestCase(TestCase):
    def test_strip_columns_from_csv(self):
        csv_content = """
            a,b,foo,c
            1,2,3,cow
            5,6,7,cookie
        """
        self.assertEqual(
            strip_columns_from_csv(csv_content, ignore_columns=["foo"]),
            "a,b,c\n1,2,cow\n5,6,cookie",
        )
        self.assertEqual(
            strip_columns_from_csv(csv_content, ignore_columns=[None]),
            "a,b,foo,c\n1,2,3,cow\n5,6,7,cookie",
        )
        self.assertEqual(
            strip_columns_from_csv(csv_content, ignore_columns="abc"), "foo\n3\n7"
        )

    def test_str_dedent(self):
        csv_content = """
            a,b,foo,c
            1,2,3,cow
            5,6,7,cookie
        """
        self.assertEqual(str_dedent(csv_content), "a,b,foo,c\n1,2,3,cow\n5,6,7,cookie")

        csv_content = """


            a,b,foo,c
            1,2,3,cow
            5,6,7,cookie


        """
        self.assertEqual(str_dedent(csv_content), "a,b,foo,c\n1,2,3,cow\n5,6,7,cookie")
        csv_content = """a,b,foo,c
            1,2,3,cow
        5,6,7,cookie


        """
        self.assertEqual(str_dedent(csv_content), "a,b,foo,c\n1,2,3,cow\n5,6,7,cookie")

        csv_content = """


            a,b,foo,c

            1,2,3,cow
            5,6,7,cookie


        """
        self.assertEqual(
            str_dedent(csv_content), "a,b,foo,c\n\n1,2,3,cow\n5,6,7,cookie"
        )


class RecipeTestCase(TestCase):
    """Test cases that can build and test a recipe"""

    maxDiff = None
    connection_string = "sqlite://"
    create_table_kwargs = {}

    def setUp(self):
        super().setUp()
        # Set up a default shelf to use
        self.shelf = self.mytable_shelf

    def assertRecipeCSV(self, recipe: Recipe, csv_text: str, ignore_columns=None):
        """Recipe data returns the supplied csv content"""
        actual = recipe.dataset.export("csv", lineterminator=str("\n")).strip("\n")
        actual = strip_columns_from_csv(actual, ignore_columns=ignore_columns)
        expected = str_dedent(csv_text).strip("\n")
        if actual != expected:
            print(f"Actual:\n{actual}\n\nExpected:\n{expected}")

        self.assertEqual(actual, expected)

    def assertRecipeSQL(self, recipe: Recipe, sql_text: str):
        """Recipe data returns the supplied csv content"""
        if str_dedent(recipe.to_sql()) != str_dedent(sql_text):
            print(f"Actual:\n{recipe.to_sql()}\n\nExpected:\n{sql_text}")
        self.assertEqual(str_dedent(recipe.to_sql()), str_dedent(sql_text))

    def assertRecipeSQLContains(self, recipe: Recipe, contains_sql_text: str):
        """Recipe data returns the supplied csv content"""
        self.assertTrue(str_dedent(contains_sql_text) in str_dedent(recipe.to_sql()))

    def assertRecipeSQLNotContains(self, recipe: Recipe, contains_sql_text: str):
        """Recipe data returns the supplied csv content"""
        self.assertTrue(
            str_dedent(contains_sql_text) not in str_dedent(recipe.to_sql())
        )

    def recipe_list(self, *args, **kwargs) -> Iterator[Recipe]:
        for potential_recipe in args:
            if isinstance(potential_recipe, Recipe):
                yield potential_recipe
            elif isinstance(potential_recipe, dict):
                yield self.recipe_from_config(potential_recipe, **kwargs)

    def recipe(self, **kwargs):
        """Construct a recipe."""
        recipe_args = {
            "shelf": getattr(self, "shelf", Shelf()),
            "session": self.session,
            "extension_classes": getattr(self, "extension_classes", []),
        }
        recipe_args.update(kwargs)
        return Recipe(**recipe_args)

    def recipe_from_config(self, config: dict, **kwargs):
        """Construct a recipe from a configuration dict."""
        recipe_args = {
            "shelf": getattr(self, "shelf", Shelf()),
            "session": self.session,
            "extension_classes": getattr(self, "extension_classes", []),
        }
        recipe_args.update(kwargs)
        shelf = recipe_args.pop("shelf")

        return Recipe.from_config(shelf, config, **recipe_args)

    @classmethod
    def load_data(cls, table_name):
        """Load data from the data/ directory"""
        table = getattr(cls, table_name)

        data = safe_load(open(os.path.join(cls.root_dir, "data", f"{table_name}.yml")))
        cls.oven.engine.execute(table.insert(), data)

    @classmethod
    def setUpClass(cls):
        """Set up tables using a connection_string to define an oven.

        Tables are loaded using data in data/{tablename}.yml
        """
        super(RecipeTestCase, cls).setUpClass()
        cls.oven = get_oven(cls.connection_string)
        cls.meta = MetaData(bind=cls.oven.engine)
        cls.session = cls.oven.Session()

        cls.root_dir = os.path.abspath(os.path.dirname(__file__))

        cls.weird_table_with_column_named_true_table = Table(
            "weird_table_with_column_named_true", cls.meta, Column("true", String)
        )

        cls.basic_table = Table(
            "foo",
            cls.meta,
            Column("first", String),
            Column("last", String),
            Column("age", Integer),
            Column("birth_date", Date),
            Column("dt", DateTime),
            **cls.create_table_kwargs,
        )

        cls.datetester_table = Table(
            "datetester", cls.meta, Column("dt", Date), Column("count", Integer)
        )

        cls.scores_table = Table(
            "scores",
            cls.meta,
            Column("username", String),
            Column("department", String),
            Column("testid", String),
            Column("score", Float),
            Column("test_date", Date),
            **cls.create_table_kwargs,
        )

        cls.datatypes_table = Table(
            "datatypes",
            cls.meta,
            Column("username", String),
            Column("department", String),
            Column("testid", String),
            Column("score", Float),
            Column("test_date", Date),
            Column("test_datetime", DateTime),
            Column("valid_score", Boolean),
            **cls.create_table_kwargs,
        )

        cls.scores_with_nulls_table = Table(
            "scores_with_nulls",
            cls.meta,
            Column("username", String),
            Column("department", String),
            Column("testid", String),
            Column("score", Float),
            Column("test_date", Date),
            **cls.create_table_kwargs,
        )

        cls.tagscores_table = Table(
            "tagscores",
            cls.meta,
            Column("username", String),
            Column("tag", String),
            Column("department", String),
            Column("testid", String),
            Column("score", Float),
            **cls.create_table_kwargs,
        )

        cls.id_tests_table = Table(
            "id_tests",
            cls.meta,
            Column("student", String),
            Column("student_id", Integer),
            Column("age", Integer),
            Column("age_id", Integer),
            Column("score", Float),
            **cls.create_table_kwargs,
        )

        cls.census_table = Table(
            "census",
            cls.meta,
            Column("state", String),
            Column("sex", String),
            Column("age", Integer),
            Column("pop2000", Integer),
            Column("pop2008", Integer),
            **cls.create_table_kwargs,
        )

        cls.state_fact_table = Table(
            "state_fact",
            cls.meta,
            Column("id", String),
            Column("name", String),
            Column("abbreviation", String),
            Column("sort", String),
            Column("status", String),
            Column("occupied", String),
            Column("notes", String),
            Column("fips_state", String),
            Column("assoc_press", String),
            Column("standard_federal_region", String),
            Column("census_region", String),
            Column("census_region_name", String),
            Column("census_division", String),
            Column("census_division_name", String),
            Column("circuit_court", String),
            **cls.create_table_kwargs,
        )

        cls.meta.drop_all(cls.oven.engine)
        cls.meta.create_all(cls.oven.engine)
        cls.load_data("weird_table_with_column_named_true_table")
        cls.load_data("basic_table")
        cls.load_data("scores_table")
        cls.load_data("datatypes_table")
        cls.load_data("scores_with_nulls_table")
        cls.load_data("tagscores_table")
        cls.load_data("id_tests_table")
        cls.load_data("census_table")
        cls.load_data("state_fact_table")

        # Load the datetester_table with dynamic date data
        start_dt = date(date.today().year, date.today().month, 1)
        data = [
            {"dt": start_dt + relativedelta(months=offset_month), "count": 1}
            for offset_month in range(-50, 50)
        ]
        cls.oven.engine.execute(cls.datetester_table.insert(), data)

        cls.mytable_shelf = Shelf(
            {
                "first": Dimension(cls.basic_table.c.first),
                "last": Dimension(cls.basic_table.c.last),
                "firstlast": Dimension(
                    cls.basic_table.c.last, id_expression=cls.basic_table.c.first
                ),
                "age": Metric(func.sum(cls.basic_table.c.age)),
            }
        )

        cls.mytable_extrarole_shelf = Shelf(
            {
                "first": Dimension(cls.basic_table.c.first),
                "last": Dimension(cls.basic_table.c.last),
                "firstlastage": Dimension(
                    cls.basic_table.c.last,
                    id_expression=cls.basic_table.c.first,
                    age_expression=cls.basic_table.c.age,
                ),
                "age": Metric(func.sum(cls.basic_table.c.age)),
            }
        )

        cls.scores_shelf = Shelf(
            {
                "username": Dimension(cls.scores_table.c.username),
                "department": Dimension(
                    cls.scores_table.c.department,
                    anonymizer=lambda value: value[::-1] if value else "None",
                ),
                "testid": Dimension(cls.scores_table.c.testid),
                "test_cnt": Metric(func.count(distinct(cls.tagscores_table.c.testid))),
                "score": Metric(func.avg(cls.scores_table.c.score)),
            }
        )

        cls.tagscores_shelf = Shelf(
            {
                "username": Dimension(cls.tagscores_table.c.username),
                "department": Dimension(cls.tagscores_table.c.department),
                "testid": Dimension(cls.tagscores_table.c.testid),
                "tag": Dimension(cls.tagscores_table.c.tag),
                "test_cnt": Metric(func.count(distinct(cls.tagscores_table.c.testid))),
                "score": Metric(
                    func.avg(cls.tagscores_table.c.score), summary_aggregation=func.sum
                ),
            }
        )

        cls.census_shelf = Shelf(
            {
                "state": Dimension(cls.census_table.c.state),
                "idvalue_state": IdValueDimension(
                    cls.census_table.c.state, "State:" + cls.census_table.c.state
                ),
                "sex": Dimension(cls.census_table.c.sex),
                "age": Dimension(cls.census_table.c.age),
                "pop2000": Metric(func.sum(cls.census_table.c.pop2000)),
                "pop2000_sum": Metric(
                    func.sum(cls.census_table.c.pop2000), summary_aggregation=func.sum
                ),
                "pop2008": Metric(func.sum(cls.census_table.c.pop2008)),
                "filter_all": Filter(1 == 0),
            }
        )

        cls.statefact_shelf = Shelf(
            {
                "state": Dimension(cls.state_fact_table.c.name),
                "abbreviation": Dimension(cls.state_fact_table.c.abbreviation),
            }
        )

    def test_sample_data_loaded(self):
        values = [
            (self.weird_table_with_column_named_true_table, 2),
            (self.basic_table, 2),
            (self.scores_table, 6),
            (self.datatypes_table, 6),
            (self.scores_with_nulls_table, 6),
            (self.tagscores_table, 10),
            (self.id_tests_table, 5),
            (self.census_table, 344),
            (self.state_fact_table, 2),
            (self.datetester_table, 100),
        ]

        for table, expected_count in values:
            self.assertEqual(
                self.session.execute(
                    select([func.count()]).select_from(table)
                ).scalar(),
                expected_count,
            )

    def test_strip_columns_from_csv(self):
        content = """a,b,c\n1,2,3"""
        c2 = strip_columns_from_csv(content, ignore_columns=["b"])
        self.assertEqual(c2, "a,c\n1,3")
